import logging
from typing import (
    Any,
    Dict,
)

from galaxy.exceptions import RequestParameterInvalidException
from galaxy.tools.parameters import visit_input_values
from galaxy.tools.parameters.basic import (
    ConnectedValue,
    contains_workflow_parameter,
    runtime_to_json,
)
from .schema import (
    AddInputAction,
    AddStepAction,
    ConnectAction,
    DisconnectAction,
    ExtractInputAction,
    ExtractUntypedParameter,
    FileDefaultsAction,
    FillStepDefaultsAction,
    InputReferenceByOrderIndex,
    OutputReferenceByOrderIndex,
    RefactorActionExecution,
    RefactorActionExecutionMessage,
    RefactorActionExecutionMessageTypeEnum,
    RefactorActions,
    RemoveUnlabeledWorkflowOutputs,
    step_reference_union,
    StepReferenceByLabel,
    UpdateAnnotationAction,
    UpdateCreatorAction,
    UpdateLicenseAction,
    UpdateNameAction,
    UpdateOutputLabelAction,
    UpdateReportAction,
    UpdateStepLabelAction,
    UpdateStepPositionAction,
    UpgradeAllStepsAction,
    UpgradeSubworkflowAction,
    UpgradeToolAction,
)
from ..modules import (
    InputParameterModule,
    NO_REPLACEMENT,
)

log = logging.getLogger(__name__)


class WorkflowRefactorExecutor:
    def __init__(self, raw_workflow_description, workflow, module_injector):
        # we mostly use the ga representation, but there may be cases where the
        # models/modules of existing workflow are more usable.
        self.raw_workflow_description = raw_workflow_description
        self.workflow = workflow
        self.module_injector = module_injector
        self.module_injector.inject_all(workflow, ignore_tool_missing_exception=True)

    def refactor(self, refactor_request: RefactorActions):
        action_executions = []
        for action in refactor_request.actions:
            # TODO: we need to regenerate a detached workflow from as_dict after
            # after each iteration here. Otherwise one set of changes might render
            # the workflow state out of sync. It is fine if you're just executing one
            # action at a time or just performing actions that use raw_workflow_description.
            action_type = action.action_type
            refactor_method_name = f"_apply_{action_type}"
            refactor_method = getattr(self, refactor_method_name, None)
            if refactor_method is None:
                raise RequestParameterInvalidException(f"Unknown workflow editing action encountered [{action_type}]")
            execution = RefactorActionExecution(
                action=action.dict(),
                messages=[],
            )
            refactor_method(action, execution)
            action_executions.append(execution)
        return action_executions

    def _apply_update_step_label(self, action: UpdateStepLabelAction, execution: RefactorActionExecution):
        step = self._find_step_for_action(action)
        step["label"] = action.label

    def _apply_update_step_position(self, action: UpdateStepPositionAction, execution: RefactorActionExecution):
        step = self._find_step_for_action(action)
        position_update = action.position_shift.to_dict()
        step["position"]["left"] = step["position"].get("left", 0) + position_update["left"]
        step["position"]["top"] = step["position"].get("top", 0) + position_update["top"]

    def _apply_update_output_label(self, action: UpdateOutputLabelAction, execution: RefactorActionExecution):
        output_reference = action.output
        output_step_dict = self._find_step(output_reference)
        output_name = output_reference.output_name
        for workflow_output_def in output_step_dict.get("workflow_outputs", []):
            if workflow_output_def["output_name"] == output_name:
                workflow_output_def["label"] = action.output_label

    def _apply_update_name(self, action: UpdateNameAction, execution: RefactorActionExecution):
        self._as_dict["name"] = action.name

    def _apply_update_annotation(self, action: UpdateAnnotationAction, execution: RefactorActionExecution):
        self._as_dict["annotation"] = action.annotation

    def _apply_update_license(self, action: UpdateLicenseAction, execution: RefactorActionExecution):
        self._as_dict["license"] = action.license

    def _apply_update_creator(self, action: UpdateCreatorAction, execution: RefactorActionExecution):
        self._as_dict["creator"] = action.creator

    def _apply_update_report(self, action: UpdateReportAction, execution: RefactorActionExecution):
        self._as_dict["report"] = {"markdown": action.report.markdown}

    def _apply_add_step(self, action: AddStepAction, execution: RefactorActionExecution):
        steps = self._as_dict["steps"]
        order_index = len(steps)
        step_dict = {
            "order_index": order_index,
            "id": "new_%d" % order_index,
            "type": action.type,
        }
        if action.tool_state:
            step_dict["tool_state"] = action.tool_state
        if action.label:
            step_dict["label"] = action.label
        if action.position:
            step_dict["position"] = action.position.to_dict()
        steps[order_index] = step_dict

    def _apply_add_input(self, action: AddInputAction, execution: RefactorActionExecution):
        input_type = action.type
        module_type = None

        tool_state: Dict[str, Any] = {}
        if input_type in ["data", "dataset"]:
            module_type = "data_input"
        elif input_type in ["data_collection", "dataset_collection"]:
            module_type = "data_collection_input"
            tool_state["collection_type"] = action.collection_type
        else:
            if input_type not in InputParameterModule.POSSIBLE_PARAMETER_TYPES:
                raise RequestParameterInvalidException(f"Invalid input type {input_type} encountered")
            module_type = "parameter_input"
            tool_state["parameter_type"] = input_type

        for action_key in ["restrictions", "suggestions", "optional", "default"]:
            value = getattr(action, action_key, None)
            if value is not None:
                tool_state[action_key] = value

        if action.restrict_on_connections is not None:
            tool_state["restrictOnConnections"] = action.restrict_on_connections

        add_step_kwds = {}
        if action.label:
            add_step_kwds["label"] = action.label

        add_step_action = AddStepAction(
            action_type="add_step", type=module_type, tool_state=tool_state, position=action.position, **add_step_kwds
        )
        self._apply_add_step(add_step_action, execution)

    def _apply_disconnect(self, action: DisconnectAction, execution: RefactorActionExecution):
        input_step_dict, input_name, output_step_dict, output_name = self._connection(action)
        output_order_index = output_step_dict["id"]  # wish this was order_index...
        # default name is name used for input's output terminal - following
        # format2 convention of allowing this be absent for clean references
        # to workflow inputs.
        all_input_connections = input_step_dict.get("input_connections")
        self.normalize_input_connections_to_list(all_input_connections, input_name)
        input_connections = all_input_connections[input_name]

        # multiple outputs attached to this inputs, just detach
        # that specific one.
        delete_index = None
        for connection_index, output in enumerate(input_connections):
            if output["id"] == output_order_index and output["output_name"] == output_name:
                delete_index = connection_index
                break
        if delete_index is None:
            raise RequestParameterInvalidException("Failed to locate connection to disconnect")
        del input_connections[delete_index]

    def _apply_connect(self, action: ConnectAction, execution: RefactorActionExecution):
        input_step_dict, input_name, output_step_dict, output_name = self._connection(action)
        output_order_index = output_step_dict["id"]  # wish this was order_index...
        all_input_connections = input_step_dict.get("input_connections")
        self.normalize_input_connections_to_list(all_input_connections, input_name, add_if_missing=True)
        input_connections = all_input_connections[input_name]
        input_connections.append(
            {
                "id": output_order_index,
                "output_name": output_name,
            }
        )

    def _apply_fill_defaults(self, action: FileDefaultsAction, execution: RefactorActionExecution):
        for _, step in self._iterate_over_step_pairs(execution):
            module = step.module
            if module.type != "tool":
                continue
            self._as_dict["steps"][step.order_index]["tool_state"] = step.module.get_tool_state()

    def _apply_fill_step_defaults(self, action: FillStepDefaultsAction, execution: RefactorActionExecution):
        step = self._find_step_with_module_for_action(action, execution)
        self._as_dict["steps"][step.order_index]["tool_state"] = step.module.get_tool_state()

    def _apply_extract_input(self, action: ExtractInputAction, execution: RefactorActionExecution):
        input_step_dict, input_name = self._input_from_action(action)
        step = self._step_with_module(input_step_dict["id"], execution)
        module = step.module
        inputs = module.get_all_inputs()

        input_def = None
        found_input_names = []
        for input in inputs:
            found_input_name = input["name"]
            found_input_names.append(found_input_name)
            if found_input_name == input_name:
                input_def = input
                break
        if input_def is None:
            raise RequestParameterInvalidException(
                f"Failed to find input with name {input_name} on step {input_step_dict['id']} - input names found {found_input_names}"
            )
        if input_def.get("multiple", False):
            raise RequestParameterInvalidException("Cannot extract input for multi-input inputs")

        module_input_type = input_def.get("input_type")
        # convert dataset, dataset_collection => data, data_collection for refactor API
        input_type = {
            "dataset": "data",
            "dataset_collection": "data_collection",
        }.get(module_input_type, module_input_type)

        input_action = AddInputAction(
            action_type="add_input",
            optional=input_def.get("optional"),
            type=input_type,
            label=action.label,
            position=action.position,
        )
        new_input_order_index = self._add_input_get_order_index(input_action, execution)
        connect_action = ConnectAction(
            action_type="connect",
            input=action.input,
            output=OutputReferenceByOrderIndex(order_index=new_input_order_index),
        )
        self._apply_connect(connect_action, execution)

    def _apply_extract_untyped_parameter(self, action: ExtractUntypedParameter, execution: RefactorActionExecution):
        untyped_parameter_name = action.name
        new_label = action.label or untyped_parameter_name
        target_value = "${%s}" % untyped_parameter_name

        target_tool_inputs = []
        rename_pjas = []

        for step_def, step in self._iterate_over_step_pairs(execution):
            module = step.module
            if module.type != "tool":
                continue

            # TODO: require a clean tool state for all tools to do this.
            tool = module.tool
            tool_inputs = module.state

            replace_tool_state = False

            def callback(input, prefixed_name, context, value=None, **kwargs):
                nonlocal replace_tool_state
                # data parameters cannot have untyped parameter values
                if input.type in ["data", "data_collection"]:
                    return NO_REPLACEMENT

                if not contains_workflow_parameter(value):
                    return NO_REPLACEMENT

                if value == target_value:
                    target_tool_inputs.append((step.order_index, input, prefixed_name))  # noqa: B023
                    replace_tool_state = True
                    return runtime_to_json(ConnectedValue())
                else:
                    return NO_REPLACEMENT

            visit_input_values(tool.inputs, tool_inputs.inputs, callback, no_replacement_value=NO_REPLACEMENT)
            if replace_tool_state:
                step_def["tool_state"] = step.module.get_tool_state()

        for post_job_action in self._iterate_over_rename_pjas():
            newname = post_job_action.get("action_arguments", {}).get("newname")
            if target_value in newname:
                rename_pjas.append(post_job_action)

        if len(target_tool_inputs) == 0 and len(rename_pjas) == 0:
            raise RequestParameterInvalidException(
                f"Failed to find {target_value} in the tool state or any workflow steps."
            )

        as_parameter_type = {
            "text": "text",
            "integer": "integer",
            "float": "float",
            "select": "text",
            "genomebuild": "text",
        }
        target_parameter_types = set()
        for _, tool_input, _ in target_tool_inputs:
            tool_input_type = tool_input.type
            if tool_input_type not in as_parameter_type:
                raise RequestParameterInvalidException(
                    "Extracting inputs for parameters on tool inputs of type {tool_input_type} is unsupported"
                )
            target_parameter_type = as_parameter_type[tool_input_type]
            target_parameter_types.add(target_parameter_type)

        if len(target_parameter_types) > 1:
            raise RequestParameterInvalidException(
                "Extracting inputs for parameters on conflicting tool input types (e.g. numeric and non-numeric) input types is unsupported"
            )

        if len(target_parameter_types) == 1:
            (target_parameter_type,) = target_parameter_types
        else:
            # only used in PJA, hence only used a string
            target_parameter_type = "text"

        for rename_pja in rename_pjas:
            # if name != label, got to rewrite this rename with new label.
            if untyped_parameter_name != new_label:
                action_arguments = rename_pja.get("action_arguments")
                old_newname = action_arguments["newname"]
                new_newname = old_newname.replace(target_value, "${%s}" % new_label)
                action_arguments["newname"] = new_newname

        optional = False
        input_action = AddInputAction(
            action_type="add_input",
            optional=optional,
            type=target_parameter_type,
            label=new_label,
            position=action.position,
        )
        new_input_order_index = self._add_input_get_order_index(input_action, execution)

        for order_index, _tool_input, prefixed_name in target_tool_inputs:
            connect_input = InputReferenceByOrderIndex(order_index=order_index, input_name=prefixed_name)
            connect_action = ConnectAction(
                action_type="connect",
                input=connect_input,
                output=OutputReferenceByOrderIndex(order_index=new_input_order_index),
            )
            self._apply_connect(connect_action, execution)

    def _apply_remove_unlabeled_workflow_outputs(
        self, action: RemoveUnlabeledWorkflowOutputs, execution: RefactorActionExecution
    ):
        for step in self._as_dict["steps"].values():
            new_outputs = []
            for workflow_output in step.get("workflow_outputs", []):
                if workflow_output.get("label") is None:
                    continue
                new_outputs.append(workflow_output)
            step["workflow_outputs"] = new_outputs

    def _apply_upgrade_subworkflow(self, action: UpgradeSubworkflowAction, execution: RefactorActionExecution):
        step_def = self._find_step(action.step)
        assert step_def["content_id"] is not None
        trans = self.module_injector.trans
        content_id = action.content_id
        if content_id is None:
            old_workflow = trans.app.workflow_manager.get_owned_workflow(trans, step_def["content_id"])
            stored_workflow = old_workflow.stored_workflow
            content_id = trans.security.encode_id(stored_workflow.latest_workflow.id)
        step_def["content_id"] = content_id
        step = self.workflow.steps[step_def["id"]]
        new_workflow = trans.app.workflow_manager.get_owned_workflow(trans, content_id)
        step.subworkflow = new_workflow
        self._inject_for_updated_step(step, execution)

        self._patch_step(execution, step, step_def)

    def _apply_upgrade_tool(self, action: UpgradeToolAction, execution: RefactorActionExecution):
        step_def = self._find_step(action.step)
        tool_id = step_def["content_id"]
        trans = self.module_injector.trans
        tool_version = action.tool_version
        if tool_version is None:
            latest_tool = trans.app.toolbox.get_tool(tool_id, get_all_versions=True)[-1]
            tool_version = latest_tool.version
            tool_id = latest_tool.id
        step = self.workflow.steps[step_def["id"]]
        step.tool_id = tool_id
        step.tool_version = tool_version
        self._inject_for_updated_step(step, execution)
        step_def["tool_version"] = tool_version
        step_def["tool_state"] = step.module.get_tool_state()
        if step_def.get("tool_id"):
            step_def["tool_id"] = step.module.get_content_id()
        if step_def.get("content_id"):
            step_def["content_id"] = step.module.get_content_id()
        self._patch_step(execution, step, step_def)

    def _apply_upgrade_all_steps(self, action: UpgradeAllStepsAction, execution: RefactorActionExecution):
        for step_order_index, step in self._as_dict["steps"].items():
            if step.get("type") == "subworkflow":
                step_action_s = UpgradeSubworkflowAction(
                    action_type="upgrade_subworkflow", step={"order_index": step_order_index}
                )
                self._apply_upgrade_subworkflow(step_action_s, execution)
            elif step.get("type") == "tool":
                step_action_t = UpgradeToolAction(action_type="upgrade_tool", step={"order_index": step_order_index})
                self._apply_upgrade_tool(step_action_t, execution)

    def _find_step(self, step_reference: step_reference_union):
        order_index = None
        if isinstance(step_reference, StepReferenceByLabel):
            label = step_reference.label
            if not label:
                raise RequestParameterInvalidException("Empty label provided.")
            for step_order_index, step in self._as_dict["steps"].items():
                if step["label"] == label:
                    order_index = step_order_index
                    break
        else:
            order_index = step_reference.order_index
        if order_index is None:
            raise RequestParameterInvalidException(f"Failed to resolve step_reference {step_reference}")
        if len(self._as_dict["steps"]) <= order_index:
            raise RequestParameterInvalidException(f"Failed to resolve step_reference {step_reference}")
        return self._as_dict["steps"][order_index]

    def _find_step_for_action(self, action):
        step_reference = action.step
        return self._find_step(step_reference)

    def _find_step_with_module_for_action(self, action, execution):
        step_reference = action.step
        step_def = self._find_step(step_reference)
        step = self.workflow.steps[step_def["id"]]
        return self._inject(step, execution)

    def _step_with_module(self, order_index, execution):
        step = self.workflow.steps[order_index]
        return self._inject(step, execution)

    def _iterate_over_step_pairs(self, execution):
        # walk over both the dict-ified steps and the model steps (ensuring)
        # module is attached.
        for order_index, step_def in self._as_dict["steps"].items():
            if order_index >= len(self.workflow.steps):
                # newly added step during refactoring, don't iterate over it...
                continue
            else:
                step = self._step_with_module(order_index, execution)
                yield step_def, step

    def _inject_for_updated_step(self, step, execution):
        step.clear_module_extras()
        return self._inject(step, execution)

    def _inject(self, step, execution):
        # compute runtime state, capture upgrade messages that result
        if not hasattr(step, "module"):
            self.module_injector.inject(step)
        self.module_injector.compute_runtime_state(step)
        if getattr(step, "upgrade_messages", None):
            for key, value in step.upgrade_messages.items():
                message = RefactorActionExecutionMessage(
                    message=value,
                    message_type=RefactorActionExecutionMessageTypeEnum.tool_state_adjustment,
                    input_name=key,
                    step_label=step.label,
                    order_index=step.order_index,
                )
                execution.messages.append(message)
        if getattr(step.module, "version_changes", None):
            for version_change in step.module.version_changes:
                message = RefactorActionExecutionMessage(
                    message=version_change,
                    message_type=RefactorActionExecutionMessageTypeEnum.tool_version_change,
                    step_label=step.label,
                    order_index=step.order_index,
                )
                execution.messages.append(message)

        return step

    def _iterate_over_rename_pjas(self):
        for _, step_def in self._as_dict["steps"].items():
            if step_def["type"] != "tool":
                continue
            post_job_actions = step_def.get("post_job_actions", [])
            for post_job_action in post_job_actions.values():
                if post_job_action["action_type"] == "RenameDatasetAction":
                    yield post_job_action

    def _add_input_get_order_index(self, input_action: AddInputAction, execution: RefactorActionExecution):
        self._apply_add_input(input_action, execution)
        return len(self._as_dict["steps"]) - 1

    def _input_from_action(self, action):
        input_reference = action.input
        input_step_dict = self._find_step(input_reference)
        input_name = input_reference.input_name
        return input_step_dict, input_name

    def _connection(self, action):
        input_step_dict, input_name = self._input_from_action(action)
        output_reference = action.output
        output_step_dict = self._find_step(output_reference)
        output_name = output_reference.output_name
        return input_step_dict, input_name, output_step_dict, output_name

    def _patch_step(self, execution, step, step_def):
        """
        patch a workflow step after upgrading the tool / subworkflow
        """
        # TODO: find workflow outputs that need to be dropped and report them
        upgrade_inputs = step.module.get_all_inputs()
        upgrade_outputs = step.module.get_all_outputs()
        upgrade_output_names = {u["name"] for u in upgrade_outputs}
        upgrade_order_index = step_def["id"]
        upgrade_label = step_def.get("label")
        all_input_connections = step_def.get("input_connections")
        inputs_to_delete = []
        for input_name, input_connections in all_input_connections.items():
            # try and find an input connection for each input
            matching_input = None
            for upgrade_input in upgrade_inputs:
                if upgrade_input["name"] == input_name:
                    matching_input = upgrade_input
                    break

            # In the future check parameter type, format, mapping status...
            if matching_input is None:
                inputs_to_delete.append(input_name)
                for input_connection in _listify_connections(input_connections):
                    message_text = (
                        f"Tool or subworkflow input '{input_name}' no longer available, dropping connection to it."
                    )
                    from_order_index = input_connection["id"]
                    from_step_label = self._as_dict["steps"].get(from_order_index, {}).get("label")
                    message = RefactorActionExecutionMessage(
                        message=message_text,
                        message_type=RefactorActionExecutionMessageTypeEnum.connection_drop_forced,
                        input_name=input_name,
                        step_label=upgrade_label,
                        order_index=upgrade_order_index,
                        output_name=input_connection["output_name"],
                        from_order_index=from_order_index,
                        from_step_label=from_step_label,
                    )
                    execution.messages.append(message)

        for input_name in inputs_to_delete:
            del all_input_connections[input_name]

        for _, step in self._as_dict["steps"].items():
            all_input_connections = step.get("input_connections")
            for input_name, input_connections in all_input_connections.items():
                rebuilt_valid_connections = []
                for input_connection in _listify_connections(input_connections):
                    include = False
                    if input_connection["id"] != upgrade_order_index:
                        include = True
                    else:
                        output_name = input_connection["output_name"]
                        if output_name in upgrade_output_names:
                            include = True
                        else:
                            # dropped outputs
                            message_text = f"Tool or subworkflow output '{output_name}' no longer available, dropping connection from it."
                            message = RefactorActionExecutionMessage(
                                message=message_text,
                                message_type=RefactorActionExecutionMessageTypeEnum.connection_drop_forced,
                                input_name=input_name,
                                step_label=step.get("label"),
                                order_index=step["id"],
                                output_name=output_name,
                                from_step_label=upgrade_label,
                                from_order_index=upgrade_order_index,
                            )
                            execution.messages.append(message)
                    if include:
                        rebuilt_valid_connections.append(input_connection)
                all_input_connections[input_name] = rebuilt_valid_connections

        workflow_outputs_to_delete = []
        for workflow_output in step_def.get("workflow_outputs", []):
            output_label = workflow_output.get("label")
            output_name = workflow_output["output_name"]
            if not output_label:
                continue

            if output_name not in upgrade_output_names:
                workflow_outputs_to_delete.append(workflow_output)
                message_text = f"Subworkflow output '{output_name}' no longer available, dropping corresponding workflow output label {output_label}."
                message = RefactorActionExecutionMessage(
                    message=message_text,
                    message_type=RefactorActionExecutionMessageTypeEnum.workflow_output_drop_forced,
                    step_label=upgrade_label,
                    order_index=upgrade_order_index,
                    output_name=workflow_output.get("output_name"),
                    output_label=output_label,
                )
                execution.messages.append(message)

    @staticmethod
    def normalize_input_connections_to_list(all_input_connections, input_name, add_if_missing=False):
        if add_if_missing and input_name not in all_input_connections:
            all_input_connections[input_name] = []
        input_connections = all_input_connections[input_name]
        all_input_connections[input_name] = _listify_connections(input_connections)

    @property
    def _as_dict(self):
        return self.raw_workflow_description.as_dict


def _listify_connections(input_connections):
    if not isinstance(input_connections, list):
        return [input_connections]
    return input_connections



import json

import requests
from social_core.actions import (
    do_auth,
    do_complete,
    do_disconnect,
)
from social_core.backends.utils import get_backend
from social_core.strategy import BaseStrategy
from social_core.utils import (
    module_member,
    setting_name,
)
from sqlalchemy.exc import IntegrityError

from galaxy.exceptions import MalformedContents
from galaxy.model import (
    PSAAssociation,
    PSACode,
    PSANonce,
    PSAPartial,
    UserAuthnzToken,
)
from galaxy.util import DEFAULT_SOCKET_TIMEOUT
from . import IdentityProvider

# key: a component name which PSA requests.
# value: is the name of a class associated with that key.
DEFAULTS = {"STRATEGY": "Strategy", "STORAGE": "Storage"}

BACKENDS = {
    "google": "social_core.backends.google_openidconnect.GoogleOpenIdConnect",
    "globus": "social_core.backends.globus.GlobusOpenIdConnect",
    "elixir": "social_core.backends.elixir.ElixirOpenIdConnect",
    "okta": "social_core.backends.okta_openidconnect.OktaOpenIdConnect",
    "azure": "social_core.backends.azuread_tenant.AzureADTenantOAuth2",
}

BACKENDS_NAME = {
    "google": "google-openidconnect",
    "globus": "globus",
    "elixir": "elixir",
    "okta": "okta-openidconnect",
    "azure": "azuread-tenant-oauth2",
}

AUTH_PIPELINE = (
    # Get the information we can about the user and return it in a simple
    # format to create the user instance later. On some cases the details are
    # already part of the auth response from the provider, but sometimes this
    # could hit a provider API.
    "social_core.pipeline.social_auth.social_details",
    # Get the social uid from whichever service we're authing thru. The uid is
    # the unique identifier of the given user in the provider.
    "social_core.pipeline.social_auth.social_uid",
    # Verifies that the current auth process is valid within the current
    # project, this is where emails and domains allowlists are applied (if
    # defined).
    "social_core.pipeline.social_auth.auth_allowed",
    # Checks if the decoded response contains all the required fields such
    # as an ID token or a refresh token.
    "galaxy.authnz.psa_authnz.contains_required_data",
    "galaxy.authnz.psa_authnz.verify",
    # Checks if the current social-account is already associated in the site.
    "social_core.pipeline.social_auth.social_user",
    # Make up a username for this person, appends a random string at the end if
    # there's any collision.
    "social_core.pipeline.user.get_username",
    # Send a validation email to the user to verify its email address.
    # 'social_core.pipeline.mail.mail_validation',
    # Associates the current social details with another user account with
    # a similar email address.
    "social_core.pipeline.social_auth.associate_by_email",
    # Create a user account if we haven't found one yet.
    "social_core.pipeline.user.create_user",
    # Create the record that associated the social account with this user.
    "social_core.pipeline.social_auth.associate_user",
    # Populate the extra_data field in the social record with the values
    # specified by settings (and the default ones like access_token, etc).
    "social_core.pipeline.social_auth.load_extra_data",
    # Update the user record with any changed info from the auth service.
    "social_core.pipeline.user.user_details",
)

DISCONNECT_PIPELINE = ("galaxy.authnz.psa_authnz.allowed_to_disconnect", "galaxy.authnz.psa_authnz.disconnect")


class PSAAuthnz(IdentityProvider):
    def __init__(self, provider, oidc_config, oidc_backend_config):
        self.config = {"provider": provider.lower()}
        for key, value in oidc_config.items():
            self.config[setting_name(key)] = value

        self.config[setting_name("USER_MODEL")] = "models.User"
        self.config["SOCIAL_AUTH_PIPELINE"] = AUTH_PIPELINE
        self.config["DISCONNECT_PIPELINE"] = DISCONNECT_PIPELINE
        self.config[setting_name("AUTHENTICATION_BACKENDS")] = (BACKENDS[provider],)

        self.config["VERIFY_SSL"] = oidc_config.get("VERIFY_SSL")
        self.config["REQUESTS_TIMEOUT"] = oidc_config.get("REQUESTS_TIMEOUT")
        self.config["ID_TOKEN_MAX_AGE"] = oidc_config.get("ID_TOKEN_MAX_AGE")

        # The following config sets PSA to call the `_login_user` function for
        # logging in a user. If this setting is set to false, the `_login_user`
        # would not be called, and as a result Galaxy would not know who is
        # the just logged-in user.
        self.config[setting_name("INACTIVE_USER_LOGIN")] = True

        if provider in BACKENDS_NAME:
            self._setup_idp(oidc_backend_config)

        # Secondary AuthZ with Google identities is currently supported
        if provider != "google":
            if "SOCIAL_AUTH_SECONDARY_AUTH_PROVIDER" in self.config:
                del self.config["SOCIAL_AUTH_SECONDARY_AUTH_PROVIDER"]
            if "SOCIAL_AUTH_SECONDARY_AUTH_ENDPOINT" in self.config:
                del self.config["SOCIAL_AUTH_SECONDARY_AUTH_ENDPOINT"]

    def _setup_idp(self, oidc_backend_config):
        self.config[setting_name("AUTH_EXTRA_ARGUMENTS")] = {"access_type": "offline"}
        self.config["KEY"] = oidc_backend_config.get("client_id")
        self.config["SECRET"] = oidc_backend_config.get("client_secret")
        self.config["redirect_uri"] = oidc_backend_config.get("redirect_uri")
        self.config["EXTRA_SCOPES"] = oidc_backend_config.get("extra_scopes")
        if oidc_backend_config.get("prompt") is not None:
            self.config[setting_name("AUTH_EXTRA_ARGUMENTS")]["prompt"] = oidc_backend_config.get("prompt")
        if oidc_backend_config.get("api_url") is not None:
            self.config[setting_name("API_URL")] = oidc_backend_config.get("api_url")
        if oidc_backend_config.get("url") is not None:
            self.config[setting_name("URL")] = oidc_backend_config.get("url")

    def _get_helper(self, name, do_import=False):
        this_config = self.config.get(setting_name(name), DEFAULTS.get(name, None))
        return do_import and module_member(this_config) or this_config

    def _load_backend(self, strategy, redirect_uri):
        backends = self._get_helper("AUTHENTICATION_BACKENDS")
        backend = get_backend(backends, BACKENDS_NAME[self.config["provider"]])
        return backend(strategy, redirect_uri)

    def _login_user(self, backend, user, social_user):
        self.config["user"] = user

    def authenticate(self, trans):
        on_the_fly_config(trans.sa_session)
        strategy = Strategy(trans.request, trans.session, Storage, self.config)
        backend = self._load_backend(strategy, self.config["redirect_uri"])
        if (
            backend.name is BACKENDS_NAME["google"]
            and "SOCIAL_AUTH_SECONDARY_AUTH_PROVIDER" in self.config
            and "SOCIAL_AUTH_SECONDARY_AUTH_ENDPOINT" in self.config
        ):
            backend.DEFAULT_SCOPE.append("https://www.googleapis.com/auth/cloud-platform")

        if self.config["EXTRA_SCOPES"] is not None:
            backend.DEFAULT_SCOPE.extend(self.config["EXTRA_SCOPES"])

        return do_auth(backend)

    def callback(self, state_token, authz_code, trans, login_redirect_url):
        on_the_fly_config(trans.sa_session)
        self.config[setting_name("LOGIN_REDIRECT_URL")] = login_redirect_url
        strategy = Strategy(trans.request, trans.session, Storage, self.config)
        strategy.session_set(f"{BACKENDS_NAME[self.config['provider']]}_state", state_token)
        backend = self._load_backend(strategy, self.config["redirect_uri"])
        redirect_url = do_complete(
            backend,
            login=lambda backend, user, social_user: self._login_user(backend, user, social_user),
            user=trans.user,
            state=state_token,
        )
        return redirect_url, self.config.get("user", None)

    def disconnect(self, provider, trans, disconnect_redirect_url=None, association_id=None):
        on_the_fly_config(trans.sa_session)
        self.config[setting_name("DISCONNECT_REDIRECT_URL")] = (
            disconnect_redirect_url if disconnect_redirect_url is not None else ()
        )
        strategy = Strategy(trans.request, trans.session, Storage, self.config)
        backend = self._load_backend(strategy, self.config["redirect_uri"])
        response = do_disconnect(backend, trans.user, association_id)
        if isinstance(response, str):
            return True, "", response
        return response.get("success", False), response.get("message", ""), ""


class Strategy(BaseStrategy):
    def __init__(self, request, session, storage, config, tpl=None):
        self.request = request
        self.session = session if session else {}
        self.config = config
        self.config["SOCIAL_AUTH_REDIRECT_IS_HTTPS"] = (
            True if self.request and self.request.host.startswith("https:") else False
        )
        self.config["SOCIAL_AUTH_GOOGLE_OPENIDCONNECT_EXTRA_DATA"] = ["id_token"]
        super().__init__(storage, tpl)

    def get_setting(self, name):
        return self.config[name]

    def session_get(self, name, default=None):
        return self.session.get(name, default)

    def session_set(self, name, value):
        self.session[name] = value

    def session_pop(self, name):
        raise NotImplementedError("Not implemented.")

    def request_data(self, merge=True):
        if not self.request:
            return {}
        if merge:
            data = self.request.GET.copy()
            data.update(self.request.POST)
        elif self.request.method == "POST":
            data = self.request.POST
        else:
            data = self.request.GET
        return data

    def request_host(self):
        if self.request:
            return self.request.host

    def build_absolute_uri(self, path=None):
        path = path or ""
        if path.startswith("http://") or path.startswith("https://"):
            return path
        if self.request:
            return (
                self.request.host + "/authnz" + ("/" + self.config.get("provider"))
                if self.config.get("provider", None) is not None
                else ""
            )
        return path

    def redirect(self, url):
        return url

    def html(self, content):
        raise NotImplementedError("Not implemented.")

    def render_html(self, tpl=None, html=None, context=None):
        raise NotImplementedError("Not implemented.")

    def start(self):
        self.clean_partial_pipeline()
        if self.backend.uses_redirect():
            return self.redirect(self.backend.auth_url())
        else:
            return self.html(self.backend.auth_html())

    def complete(self, *args, **kwargs):
        return self.backend.auth_complete(*args, **kwargs)

    def continue_pipeline(self, *args, **kwargs):
        return self.backend.continue_pipeline(*args, **kwargs)


class Storage:
    user = UserAuthnzToken
    nonce = PSANonce
    association = PSAAssociation
    code = PSACode
    partial = PSAPartial

    @classmethod
    def is_integrity_error(cls, exception):
        return exception.__class__ is IntegrityError


def on_the_fly_config(sa_session):
    PSACode.sa_session = sa_session
    UserAuthnzToken.sa_session = sa_session
    PSANonce.sa_session = sa_session
    PSAPartial.sa_session = sa_session
    PSAAssociation.sa_session = sa_session


def contains_required_data(response=None, is_new=False, **kwargs):
    """
    This function is called as part of authentication and authorization
    pipeline before user is authenticated or authorized (see AUTH_PIPELINE).

    This function asserts if all the data required by Galaxy for a user
    is provided. It raises an exception if any of the required data is missing,
    and returns void if otherwise.

    :type  response: dict
    :param response:  a dictionary containing decoded response from
                      OIDC backend that contain the following keys
                      among others:

                        -   id_token;       see: http://openid.net/specs/openid-connect-core-1_0.html#IDToken
                        -   access_token;   see: https://tools.ietf.org/html/rfc6749#section-1.4
                        -   refresh_token;  see: https://tools.ietf.org/html/rfc6749#section-1.5
                        -   token_type;     see: https://tools.ietf.org/html/rfc6750#section-6.1.1
                        -   scope;          see: http://openid.net/specs/openid-connect-core-1_0.html#AuthRequest
                        -   expires_in;     is the expiration time of the access and ID tokens in seconds since
                                            the response was generated.

    :type  is_new: bool
    :param is_new: has the user been authenticated?

    :param kwargs:      may contain the following keys among others:

                        -   uid:        user ID
                        -   user:       Galaxy user; if user is already authenticated
                        -   backend:    the backend that is used for user authentication.
                        -   storage:    an instance of Storage class.
                        -   strategy:   an instance of the Strategy class.
                        -   state:      the state code received from identity provider.
                        -   details:    details about the user's third-party identity as requested in `scope`.

    :rtype:  void
    :return: Raises an exception if any of the required arguments is missing, and pass if all are given.
    """
    hint_msg = (
        "Visit the identity provider's permitted applications page "
        "(e.g., visit `https://myaccount.google.com/u/0/permissions` "
        "for Google), then revoke the access of this Galaxy instance, "
        "and then retry to login. If the problem persists, contact "
        "the Admin of this Galaxy instance."
    )
    if response is None or not isinstance(response, dict):
        # This can happen only if PSA is not able to decode the `authnz code`
        # sent back from the identity provider. PSA internally handles such
        # scenarios; however, this case is implemented to prevent uncaught
        # server-side errors.
        raise MalformedContents(err_msg=f"`response` not found. {hint_msg}")
    if not response.get("id_token"):
        # This can happen if a non-OIDC compliant backend is used;
        # e.g., an OAuth2.0-based backend that only generates access token.
        raise MalformedContents(err_msg=f"Missing identity token. {hint_msg}")
    if is_new and not response.get("refresh_token"):
        # An identity provider (e.g., Google) sends a refresh token the first
        # time user consents Galaxy's access (i.e., the first time user logs in
        # to a galaxy instance using their credentials with the identity provider).
        # There could be variety of scenarios under which a refresh token might
        # be missing; e.g., a manipulated Galaxy's database, where a user's records
        # from galaxy_user and oidc_user_authnz_tokens tables deleted after the
        # user has provided consent. This can also happen under dev efforts.
        # The solution is to revoke the consent by visiting the identity provider's
        # website, and then retry the login process.
        raise MalformedContents(err_msg=f"Missing refresh token. {hint_msg}")


def verify(strategy=None, response=None, details=None, **kwargs):
    provider = strategy.config.get("SOCIAL_AUTH_SECONDARY_AUTH_PROVIDER")
    endpoint = strategy.config.get("SOCIAL_AUTH_SECONDARY_AUTH_ENDPOINT")
    if provider is None or endpoint is None:
        # Either the secondary authorization is not configured or OIDC IdP
        # is not compatible, so allow user login.
        return

    if provider.lower() == "gcp":
        result = requests.post(
            f"https://iam.googleapis.com/v1/projects/-/serviceAccounts/{endpoint}:getIamPolicy",
            headers={
                "Authorization": f"Bearer {response.get('access_token')}",
                "Accept": "application/json",
            },
            timeout=DEFAULT_SOCKET_TIMEOUT,
        )
        res = json.loads(result.content)
        if result.status_code == requests.codes.ok:
            email_addresses = res["bindings"][0]["members"]
            email_addresses = [x.lower().replace("user:", "").strip() for x in email_addresses]
            if details.get("email") in email_addresses:
                # Secondary authorization successful, so allow user login.
                pass
            else:
                raise Exception("Not authorized by GCP IAM.")
        else:
            # The message of the raised exception is shown to the user; hence,
            # the following way of handling exception is better than using
            # result.raise_for_status(), since raise_for_status may report
            # sensitive information that should not be exposed to users.
            raise Exception(res["error"]["message"])
    else:
        raise Exception(f"`{provider}` is an unsupported secondary authorization provider, contact admin.")


def allowed_to_disconnect(
    name=None, user=None, user_storage=None, strategy=None, backend=None, request=None, details=None, **kwargs
):
    """
    Disconnect is the process of disassociating a Galaxy user and a third-party authnz.
    In other words, it is the process of removing any access and/or ID tokens of a user.
    This function should raise an exception if disconnection is NOT permitted. Do NOT
    return any value (except an empty dictionary) if disconnect is allowed. Because, at
    least until PSA social_core v.1.5.0, any returned value (e.g., Boolean) will result
    in ignoring the rest of the disconnect pipeline.
    See the following condition in `run_pipeline` function:
    https://github.com/python-social-auth/social-core/blob/master/social_core/backends/base.py#L114
    :param name: name of the backend (e.g., google-openidconnect)
    :type user: galaxy.model.User
    :type user_storage: galaxy.model.UserAuthnzToken
    :type strategy: galaxy.authnz.psa_authnz.Strategy
    :type backend: PSA backend object (e.g., social_core.backends.google_openidconnect.GoogleOpenIdConnect)
    :type request: webob.multidict.MultiDict
    :type details: dict
    :return: empty dict
    """


def disconnect(
    name=None, user=None, user_storage=None, strategy=None, backend=None, request=None, details=None, **kwargs
):
    """
    Disconnect is the process of disassociating a Galaxy user and a third-party authnz.
    In other words, it is the process of removing any access and/or ID tokens of a user.
    :param name: name of the backend (e.g., google-openidconnect)
    :type user: galaxy.model.User
    :type user_storage: galaxy.model.UserAuthnzToken
    :type strategy: galaxy.authnz.psa_authnz.Strategy
    :type backend: PSA backend object (e.g., social_core.backends.google_openidconnect.GoogleOpenIdConnect)
    :type request: webob.multidict.MultiDict
    :type details: dict
    :return: void or empty dict. Any key-value pair inside the dictionary will be available
    inside PSA only, and will be passed to the next step in the disconnect pipeline. However,
    the key-value pair will not be returned as a result of calling the `do_disconnect` function.
    Additionally, returning any value except for a(n) (empty) dictionary, will break the
    disconnect pipeline, and that value will be returned as a result of calling the `do_disconnect` function.
    """

    sa_session = user_storage.sa_session
    user_authnz = (
        sa_session.query(user_storage)
        .filter(user_storage.table.c.user_id == user.id, user_storage.table.c.provider == name)
        .first()
    )
    if user_authnz is None:
        return {"success": False, "message": "Not authenticated by any identity providers."}
    # option A
    sa_session.delete(user_authnz)
    # option B
    # user_authnz.extra_data = None
    sa_session.flush()



"""Manager and Serializer for library datasets."""
import logging

from galaxy import (
    model,
    util,
)
from galaxy.exceptions import (
    InsufficientPermissionsException,
    InternalServerError,
    ObjectNotFound,
    RequestParameterInvalidException,
)
from galaxy.managers import datasets
from galaxy.model import tags
from galaxy.structured_app import MinimalManagerApp
from galaxy.util import validation

log = logging.getLogger(__name__)


class LibraryDatasetsManager(datasets.DatasetAssociationManager):
    """Interface/service object for interacting with library datasets."""

    model_class = model.LibraryDatasetDatasetAssociation

    def __init__(self, app: MinimalManagerApp):
        self.app = app
        self.tag_handler = tags.GalaxyTagHandler(app.model.context)

    def get(self, trans, decoded_library_dataset_id, check_accessible=True):
        """
        Get the library dataset from the DB.

        :param  decoded_library_dataset_id: decoded library dataset id
        :type   decoded_library_dataset_id: int
        :param  check_accessible:           flag whether to check that user can access item
        :type   check_accessible:           bool

        :returns:   the requested library dataset
        :rtype:     galaxy.model.LibraryDataset
        """
        try:
            ld = (
                trans.sa_session.query(trans.app.model.LibraryDataset)
                .filter(trans.app.model.LibraryDataset.table.c.id == decoded_library_dataset_id)
                .one()
            )
        except Exception as e:
            raise InternalServerError(f"Error loading from the database.{util.unicodify(e)}")
        ld = self.secure(trans, ld, check_accessible)
        return ld

    def update(self, trans, ld, payload):
        """
        Update the given library dataset - the latest linked ldda.
        Updating older lddas (versions) is not allowed.

        :param  ld:                 library dataset to change
        :type   ld:                 LibraryDataset
        :param  payload:            dictionary structure containing::
            :param name:            new ld's name, must be longer than 0
            :type  name:            str
            :param misc_info:       new ld's misc info
            :type  misc_info:       str
            :param file_ext:        new ld's extension, must exist in the Galaxy registry
            :type  file_ext:        str
            :param genome_build:    new ld's genome build
            :type  genome_build:    str
            :param tags:            list of dataset tags
            :type  tags:            list
        :type   payload: dict

        :returns:   the changed library dataset
        :rtype:     galaxy.model.LibraryDataset
        """
        self.check_modifiable(trans, ld)
        # we are going to operate on the actual latest ldda
        ldda = ld.library_dataset_dataset_association
        payload = self._validate_and_parse_update_payload(payload)
        self._set_from_dict(trans, ldda, payload)
        return ld

    def _set_from_dict(self, trans, ldda, new_data):
        changed = False
        new_name = new_data.get("name", None)
        if new_name is not None and new_name != ldda.name:
            ldda.name = new_name
            changed = True
        new_misc_info = new_data.get("misc_info", None)
        if new_misc_info is not None and new_misc_info != ldda.info:
            ldda.info = new_misc_info
            changed = True
        new_message = new_data.get("message", None)
        if new_message is not None and new_message != ldda.message:
            ldda.message = new_message
            changed = True
        new_file_ext = new_data.get("file_ext", None)
        if new_file_ext == "auto":
            self.detect_datatype(trans, ldda)
        elif new_file_ext is not None and new_file_ext != ldda.extension:
            ldda.extension = new_file_ext
            self.set_metadata(trans, ldda)
            changed = True
        new_genome_build = new_data.get("genome_build", None)
        if new_genome_build is not None and new_genome_build != ldda.dbkey:
            ldda.dbkey = new_genome_build
            changed = True
        new_tags = new_data.get("tags", None)
        if new_tags is not None and new_tags != ldda.tags:
            self.tag_handler.delete_item_tags(item=ldda, user=trans.user)
            tag_list = self.tag_handler.parse_tags_list(new_tags)
            for tag in tag_list:
                self.tag_handler.apply_item_tag(item=ldda, user=trans.user, name=tag[0], value=tag[1])
            changed = True
        if changed:
            ldda.update_parent_folder_update_times()
            trans.sa_session.add(ldda)
            trans.sa_session.flush()
        return changed

    def _validate_and_parse_update_payload(self, payload):
        MINIMUM_STRING_LENGTH = 1
        validated_payload = {}
        for key, val in payload.items():
            if val is None:
                continue
            if key in ("name"):
                if len(val) < MINIMUM_STRING_LENGTH:
                    raise RequestParameterInvalidException(
                        f"{key} must have at least length of {MINIMUM_STRING_LENGTH}"
                    )
                val = validation.validate_and_sanitize_basestring(key, val)
                validated_payload[key] = val
            if key in ("misc_info", "message"):
                val = validation.validate_and_sanitize_basestring(key, val)
                validated_payload[key] = val
            if key in ("file_ext"):
                datatype = self.app.datatypes_registry.get_datatype_by_extension(val)
                if datatype is None and val not in ("auto",):
                    raise RequestParameterInvalidException(f"This Galaxy does not recognize the datatype of: {val}")
                validated_payload[key] = val
            if key in ("genome_build"):
                if len(val) < MINIMUM_STRING_LENGTH:
                    raise RequestParameterInvalidException(
                        f"{key} must have at least length of {MINIMUM_STRING_LENGTH}"
                    )
                val = validation.validate_and_sanitize_basestring(key, val)
                validated_payload[key] = val
            if key in ("tags"):
                val = validation.validate_and_sanitize_basestring_list(key, util.listify(val))
                validated_payload[key] = val
        return validated_payload

    def secure(self, trans, ld, check_accessible=True, check_ownership=False):
        """
        Check if library dataset is accessible to current user or the user is an admin.

        :param  ld:         library dataset
        :type   ld:         galaxy.model.LibraryDataset
        :param  check_accessible:        flag whether to check that user can access library dataset
        :type   check_accessible:        bool

        :returns:   the original library dataset
        :rtype:     galaxy.model.LibraryDataset
        """
        if trans.user_is_admin:
            # all operations are available to an admin
            return ld
        if check_accessible:
            ld = self.check_accessible(trans, ld)
        return ld

    def check_accessible(self, trans, ld):
        """
        Check whether the current user has permissions to access library dataset.

        :param  ld: library dataset
        :type   ld: galaxy.model.LibraryDataset

        :returns:   the original library dataset
        :rtype:     galaxy.model.LibraryDataset

        :raises:    ObjectNotFound
        """
        if not trans.app.security_agent.can_access_library_item(trans.get_current_user_roles(), ld, trans.user):
            raise ObjectNotFound("Library dataset with the id provided was not found.")
        elif ld.deleted:
            raise ObjectNotFound("Library dataset with the id provided is deleted.")
        else:
            return ld

    def check_modifiable(self, trans, ld):
        """
        Check whether the current user has permissions to modify library dataset.

        :param  ld: library dataset
        :type   ld: galaxy.model.LibraryDataset

        :returns:   the original library dataset
        :rtype:     galaxy.model.LibraryDataset

        :raises:    ObjectNotFound
        """
        if ld.deleted:
            raise ObjectNotFound("Library dataset with the id provided is deleted.")
        elif trans.user_is_admin:
            return ld
        if not trans.app.security_agent.can_modify_library_item(trans.get_current_user_roles(), ld):
            raise InsufficientPermissionsException("You do not have proper permission to modify this library dataset.")
        else:
            return ld

    def serialize(self, trans, ld):
        """Serialize the library dataset into a dictionary."""
        current_user_roles = trans.get_current_user_roles()

        # Build the full path for breadcrumb purposes.
        full_path = self._build_path(trans, ld.folder)
        dataset_item = (trans.security.encode_id(ld.id), ld.name)
        full_path.insert(0, dataset_item)
        full_path = full_path[::-1]

        # Find expired versions of the library dataset
        expired_ldda_versions = []
        for expired_ldda in ld.expired_datasets:
            expired_ldda_versions.append((trans.security.encode_id(expired_ldda.id), expired_ldda.name))

        rval = trans.security.encode_all_ids(ld.to_dict())
        if len(expired_ldda_versions) > 0:
            rval["has_versions"] = True
            rval["expired_versions"] = expired_ldda_versions

        ldda = ld.library_dataset_dataset_association
        if ldda.creating_job_associations:
            if ldda.creating_job_associations[0].job.stdout:
                rval["job_stdout"] = ldda.creating_job_associations[0].job.stdout.strip()
            if ldda.creating_job_associations[0].job.stderr:
                rval["job_stderr"] = ldda.creating_job_associations[0].job.stderr.strip()
        if ldda.dataset.uuid:
            rval["uuid"] = str(ldda.dataset.uuid)
        rval["deleted"] = ld.deleted
        rval["folder_id"] = f"F{rval['folder_id']}"
        rval["full_path"] = full_path
        rval["file_size"] = util.nice_size(int(ldda.get_size(calculate_size=False)))
        rval["date_uploaded"] = ldda.create_time.isoformat()
        rval["update_time"] = ldda.update_time.isoformat()
        rval["can_user_modify"] = trans.user_is_admin or trans.app.security_agent.can_modify_library_item(
            current_user_roles, ld
        )
        rval["is_unrestricted"] = trans.app.security_agent.dataset_is_public(ldda.dataset)
        rval["tags"] = self.tag_handler.get_tags_str(ldda.tags)

        #  Manage dataset permission is always attached to the dataset itself, not the the ld or ldda to maintain consistency
        rval["can_user_manage"] = trans.user_is_admin or trans.app.security_agent.can_manage_dataset(
            current_user_roles, ldda.dataset
        )
        return rval

    def _build_path(self, trans, folder):
        """
        Search the path upwards recursively and load the whole route of
        names and ids for breadcrumb building purposes.

        :param folder: current folder for navigating up
        :param type:   Galaxy LibraryFolder

        :returns:   list consisting of full path to the library
        :type:      list
        """
        path_to_root = []
        if folder.parent_id is None:
            # We are almost in root
            path_to_root.append((f"F{trans.security.encode_id(folder.id)}", folder.name))
        else:
            # We add the current folder and traverse up one folder.
            path_to_root.append((f"F{trans.security.encode_id(folder.id)}", folder.name))
            upper_folder = trans.sa_session.query(trans.app.model.LibraryFolder).get(folder.parent_id)
            path_to_root.extend(self._build_path(trans, upper_folder))
        return path_to_root



""" Module containing Galaxy workflow scheduling plugins. Galaxy's interface
for workflow scheduling is highly experimental and the interface required for
scheduling plugins will almost certainly change.
"""
from abc import (
    ABCMeta,
    abstractmethod,
)


class WorkflowSchedulingPlugin(metaclass=ABCMeta):
    """A plugin defining how Galaxy should schedule plugins. By default
    plugins are passive and should monitor Galaxy's work queue for
    WorkflowRequests. Inherit from ActiveWorkflowSchedulingPlugin instead if
    the scheduling plugin should be forced (i.e. if scheduling happen all at
    once or the request will be stored and monitored outside of Galaxy.)
    """

    @property
    @abstractmethod
    def plugin_type(self):
        """Short string providing labelling this plugin"""

    @abstractmethod
    def startup(self, app):
        """Called when Galaxy starts up if the plugin is enabled."""

    @abstractmethod
    def shutdown(self):
        """Called when Galaxy is shutting down, workflow scheduling should
        end.
        """


class ActiveWorkflowSchedulingPlugin(WorkflowSchedulingPlugin, metaclass=ABCMeta):
    @abstractmethod
    def schedule(self, workflow_invocation):
        """Optionally return one or more commands to instrument job. These
        commands will be executed on the compute server prior to the job
        running.
        """



"""
Manager and Serializer for Roles.
"""
import logging
from typing import List

from sqlalchemy import false
from sqlalchemy.orm import exc as sqlalchemy_exceptions

import galaxy.exceptions
from galaxy import model
from galaxy.exceptions import RequestParameterInvalidException
from galaxy.managers import base
from galaxy.managers.context import ProvidesUserContext
from galaxy.model import Role
from galaxy.schema.schema import RoleDefinitionModel
from galaxy.util import unicodify

log = logging.getLogger(__name__)


class RoleManager(base.ModelManager[model.Role]):
    """
    Business logic for roles.
    """

    model_class = model.Role
    foreign_key_name = "role"

    user_assoc = model.UserRoleAssociation
    group_assoc = model.GroupRoleAssociation

    def get(self, trans: ProvidesUserContext, role_id: int) -> model.Role:
        """
        Method loads the role from the DB based on the given role id.

        :param  role_id:      id of the role to load from the DB
        :type   role_id:      int

        :returns:   the loaded Role object
        :rtype:     galaxy.model.Role

        :raises: InconsistentDatabase, RequestParameterInvalidException, InternalServerError
        """
        try:
            role = self.session().query(self.model_class).filter(self.model_class.id == role_id).one()
        except sqlalchemy_exceptions.MultipleResultsFound:
            raise galaxy.exceptions.InconsistentDatabase("Multiple roles found with the same id.")
        except sqlalchemy_exceptions.NoResultFound:
            raise galaxy.exceptions.RequestParameterInvalidException("No accessible role found with the id provided.")
        except Exception as e:
            raise galaxy.exceptions.InternalServerError(f"Error loading from the database.{unicodify(e)}")

        if not (trans.user_is_admin or trans.app.security_agent.ok_to_display(trans.user, role)):
            raise galaxy.exceptions.RequestParameterInvalidException("No accessible role found with the id provided.")

        return role

    def list_displayable_roles(self, trans: ProvidesUserContext) -> List[Role]:
        roles = []
        for role in trans.sa_session.query(Role).filter(Role.deleted == false()):
            if trans.user_is_admin or trans.app.security_agent.ok_to_display(trans.user, role):
                roles.append(role)
        return roles

    def create_role(self, trans: ProvidesUserContext, role_definition_model: RoleDefinitionModel) -> model.Role:
        name = role_definition_model.name
        description = role_definition_model.description
        user_ids = role_definition_model.user_ids or []
        group_ids = role_definition_model.group_ids or []

        if trans.sa_session.query(Role).filter(Role.name == name).first():
            raise RequestParameterInvalidException(f"A role with that name already exists [{name}]")

        role_type = Role.types.ADMIN  # TODO: allow non-admins to create roles

        role = Role(name=name, description=description, type=role_type)
        trans.sa_session.add(role)
        users = [trans.sa_session.query(model.User).get(i) for i in user_ids]
        groups = [trans.sa_session.query(model.Group).get(i) for i in group_ids]

        # Create the UserRoleAssociations
        for user in users:
            trans.app.security_agent.associate_user_role(user, role)

        # Create the GroupRoleAssociations
        for group in groups:
            trans.app.security_agent.associate_group_role(group, role)

        trans.sa_session.flush()
        return role



#
# Filters a GFF file using a list of attribute values. Attribute values must
# be in the first column of the file; subsequent columns are ignored.
# Usage:
# python gff_filter_by_attribute_values.py <gff_file> <attribute_name> <ids_file> <output_file>
#
from __future__ import print_function

import sys


def parse_gff_attributes(attr_str):
    """
    Parses a GFF/GTF attribute string and returns a dictionary of name-value
    pairs. The general format for a GFF3 attributes string is
        name1=value1;name2=value2
    The general format for a GTF attribute string is
        name1 "value1" ; name2 "value2"
    The general format for a GFF attribute string is a single string that
    denotes the interval's group; in this case, method returns a dictionary
    with a single key-value pair, and key name is 'group'
    """
    attributes_list = attr_str.split(";")
    attributes = {}
    for name_value_pair in attributes_list:
        # Try splitting by space and, if necessary, by '=' sign.
        pair = name_value_pair.strip().split(" ")
        if len(pair) == 1:
            pair = name_value_pair.strip().split("=")
        if len(pair) == 1:
            # Could not split for some reason -- raise exception?
            continue
        if pair == "":
            continue
        name = pair[0].strip()
        if name == "":
            continue
        # Need to strip double quote from values
        value = pair[1].strip(' "')
        attributes[name] = value

    if len(attributes) == 0:
        # Could not split attributes string, so entire string must be
        # 'group' attribute. This is the case for strictly GFF files.
        attributes["group"] = attr_str
    return attributes


def gff_filter(gff_file, attribute_name, ids_file, output_file):
    # Put ids in dict for quick lookup.
    ids_dict = {}
    for line in open(ids_file):
        ids_dict[line.split("\t")[0].strip()] = True

    # Filter GFF file using ids.
    with open(output_file, "w") as output, open(gff_file) as ingff:
        for line in ingff:
            if not line or line.startswith("#"):
                output.write(line)
                continue
            fields = line.split("\t")
            attributes = parse_gff_attributes(fields[8])
            if attribute_name in attributes and attributes[attribute_name] in ids_dict:
                output.write(line)


if __name__ == "__main__":
    # Handle args.
    if len(sys.argv) != 5:
        print("usage: python %s <gff_file> <attribute_name> <ids_file> <output_file>" % sys.argv[0], file=sys.stderr)
        sys.exit(-1)
    gff_file, attribute_name, ids_file, output_file = sys.argv[1:]
    gff_filter(gff_file, attribute_name, ids_file, output_file)



"""The module describes the ``sentry`` error plugin plugin."""
import logging

try:
    import sentry_sdk
except ImportError:
    sentry_sdk = None

from galaxy import web
from galaxy.util import string_as_bool
from . import ErrorPlugin

log = logging.getLogger(__name__)

SENTRY_SDK_IMPORT_MESSAGE = "The Python sentry-sdk package is required to use this feature, please install it"
ERROR_TEMPLATE = """Galaxy Job Error: {tool_id} v{tool_version}

Command Line:
{command_line}

Stderr:
{stderr}

Stdout:
{stdout}

The user provided the following information:
{message}"""


class SentryPlugin(ErrorPlugin):
    """Send error report to Sentry."""

    plugin_type = "sentry"

    def __init__(self, **kwargs):
        self.app = kwargs["app"]
        self.redact_user_details_in_bugreport = self.app.config.redact_user_details_in_bugreport
        self.verbose = string_as_bool(kwargs.get("verbose", False))
        self.user_submission = string_as_bool(kwargs.get("user_submission", False))
        assert sentry_sdk, SENTRY_SDK_IMPORT_MESSAGE

    def submit_report(self, dataset, job, tool, **kwargs):
        """Submit the error report to sentry"""
        extra = {
            "info": job.info,
            "id": job.id,
            "command_line": job.command_line,
            "destination_id": job.destination_id,
            "stderr": job.stderr,
            "traceback": job.traceback,
            "exit_code": job.exit_code,
            "stdout": job.stdout,
            "handler": job.handler,
            "tool_id": job.tool_id,
            "tool_version": job.tool_version,
            "tool_xml": tool.config_file if tool else None,
        }
        if self.redact_user_details_in_bugreport:
            extra["email"] = "redacted"
        else:
            if "email" in kwargs:
                extra["email"] = kwargs["email"]

        # User submitted message
        extra["message"] = kwargs.get("message", "")

        # Construct the error message to send to sentry. The first line
        # will be the issue title, everything after that becomes the
        # "message"
        error_message = ERROR_TEMPLATE.format(**extra)

        # Update context with user information in a sentry-specific manner
        context = {}

        # Getting the url allows us to link to the dataset info page in case
        # anything is missing from this report.
        try:
            url = web.url_for(
                controller="dataset",
                action="show_params",
                dataset_id=self.app.security.encode_id(dataset.id),
                qualified=True,
            )
        except AttributeError:
            # The above does not work when handlers are separate from the web handlers
            url = None

        user = job.get_user()
        if self.redact_user_details_in_bugreport:
            if user:
                # Opaque identifier
                context["user"] = {"id": user.id}
        else:
            if user:
                # User information here also places email links + allows seeing
                # a list of affected users in the tags/filtering.
                context["user"] = {
                    "name": user.username,
                    "email": user.email,
                }

        context["request"] = {"url": url}

        for key, value in context.items():
            sentry_sdk.set_context(key, value)
        sentry_sdk.set_context("job", extra)
        sentry_sdk.set_tag("tool_id", job.tool_id)
        sentry_sdk.set_tag("tool_version", job.tool_version)

        # Send the message, using message because
        response = sentry_sdk.capture_message(error_message)
        return (f"Submitted bug report to Sentry. Your guru meditation number is {response}", "success")


__all__ = ("SentryPlugin",)



from pathlib import Path
from typing import (
    cast,
    Dict,
    Optional,
)

from galaxy import exceptions
from galaxy.files import ConfiguredFileSources
from galaxy.files.uris import stream_url_to_file
from galaxy.model import DatasetInstance
from galaxy.structured_app import (
    MinimalManagerApp,
    StructuredApp,
)
from galaxy.tool_util.data import BundleProcessingOptions
from galaxy.tool_util.data._schema import (
    ToolDataDetails,
    ToolDataEntryList,
    ToolDataField,
)
from galaxy.tools.data import (
    TabularToolDataField,
    TabularToolDataTable,
    ToolDataTable,
    ToolDataTableManager,
)


class ToolDataManager:
    """
    Interface/service object for interacting with tool data.
    """

    def __init__(self, app: StructuredApp):
        self._app = app

    @property
    def data_tables(self) -> Dict[str, ToolDataTable]:
        return self._app.tool_data_tables.data_tables

    def index(self) -> ToolDataEntryList:
        """Return all tool data tables."""
        return self._app.tool_data_tables.index()

    def show(self, table_name: str) -> ToolDataDetails:
        """Get details of a given data table"""
        data_table = self._data_table(table_name)
        element_view = data_table.to_dict(view="element")
        return ToolDataDetails.construct(**element_view)

    def show_field(self, table_name: str, field_name: str) -> ToolDataField:
        """Get information about a partiular field in a tool data table"""
        field = self._data_table_field(table_name, field_name)
        return ToolDataField.construct(**field.to_dict())

    def reload(self, table_name: str) -> ToolDataDetails:
        """Reloads a tool data table."""
        data_table = self._data_table(table_name)
        data_table.reload_from_files()
        return self._reload_data_table(table_name)

    def get_field_file_path(self, table_name: str, field_name: str, file_name: str) -> Path:
        """Get the absolute path to a given file name in the table field"""
        field_value = self._data_table_field(table_name, field_name)
        base_dir = Path(field_value.get_base_dir())
        full_path = base_dir / file_name
        if str(full_path) not in field_value.get_files():
            raise exceptions.ObjectNotFound("No such path in data table field.")
        return full_path.absolute()

    def delete(self, table_name: str, values: Optional[str] = None) -> ToolDataDetails:
        """Removes an item from a data table"""
        data_table = self._tabular_data_table(table_name)
        if not values:
            raise exceptions.RequestParameterInvalidException("Invalid values for data table item specified.")

        split_values = values.split("\t")

        if len(split_values) != len(data_table.get_column_name_list()):
            raise exceptions.RequestParameterInvalidException(
                f"Invalid data table item ( {values} ) specified. Wrong number of columns ({len(split_values)} given, {len(data_table.get_column_name_list())} required)."
            )

        data_table.remove_entry(split_values)
        return self._reload_data_table(table_name)

    def _tabular_data_table(self, table_name: str) -> TabularToolDataTable:
        return cast(TabularToolDataTable, self._data_table(table_name))

    def _data_table(self, table_name: str) -> ToolDataTable:
        try:
            return self.data_tables[table_name]
        except KeyError:
            raise exceptions.ObjectNotFound(f"No such data table {table_name}")

    def _data_table_field(self, table_name: str, field_name: str) -> TabularToolDataField:
        out = self._tabular_data_table(table_name).get_field(field_name)
        if out is None:
            raise exceptions.ObjectNotFound(f"No such field {field_name} in data table {table_name}.")
        return out

    def _reload_data_table(self, name: str) -> ToolDataDetails:
        self._app.queue_worker.send_control_task("reload_tool_data_tables", noop_self=True, kwargs={"table_name": name})
        return self.show(name)


class ToolDataImportManager:
    file_sources: ConfiguredFileSources
    tool_data_tables: ToolDataTableManager

    def __init__(self, app: MinimalManagerApp):
        self.file_sources = app.file_sources
        self.tool_data_tables = app.tool_data_tables

    def import_data_bundle_by_uri(self, config, uri: str, tool_data_file_path=None):
        # an admin-only task - so allow file:// uris
        if uri.startswith("file://"):
            target = uri[len("file://") :]
        else:
            target = stream_url_to_file(
                uri,
                self.file_sources,
            )
        options = BundleProcessingOptions(
            what="data import",  # An alternative to this is sticking this in the bundle, only used for logging.
            data_manager_path=config.galaxy_data_manager_data_path,
            target_config_file=config.data_manager_config_file,
            tool_data_file_path=tool_data_file_path,
        )
        self.tool_data_tables.import_bundle(
            target,
            options,
        )

    def import_data_bundle_by_dataset(self, config, dataset: DatasetInstance, tool_data_file_path=None):
        options = BundleProcessingOptions(
            what="data import",  # An alternative to this is sticking this in the bundle, only used for logging.
            data_manager_path=config.galaxy_data_manager_data_path,
            target_config_file=config.data_manager_config_file,
            tool_data_file_path=tool_data_file_path,
        )
        self.tool_data_tables.import_bundle(
            dataset.extra_files_path,
            options,
        )



import getpass
import logging
import os
import shutil

from galaxy import model
from galaxy.model import store
from galaxy.schema.tasks import SetupHistoryExportJob
from galaxy.util.path import external_chown

log = logging.getLogger(__name__)


class JobImportHistoryArchiveWrapper:
    """
    Class provides support for performing jobs that import a history from
    an archive.
    """

    def __init__(self, app, job_id):
        self.app = app
        self.job_id = job_id
        self.sa_session = self.app.model.context

    def setup_job(self, jiha, archive_source, archive_type):
        if self.app.config.external_chown_script:
            if archive_type != "url":
                external_chown(
                    archive_source,
                    jiha.job.user.system_user_pwent(self.app.config.real_system_username),
                    self.app.config.external_chown_script,
                    "history import archive",
                )
            external_chown(
                jiha.archive_dir,
                jiha.job.user.system_user_pwent(self.app.config.real_system_username),
                self.app.config.external_chown_script,
                "history import archive directory",
            )

    def cleanup_after_job(self):
        """Set history, datasets, collections and jobs' attributes
        and clean up archive directory.
        """

        #
        # Import history.
        #

        jiha = self.sa_session.query(model.JobImportHistoryArchive).filter_by(job_id=self.job_id).first()
        if not jiha:
            return None
        user = jiha.job.user

        new_history = None
        try:
            archive_dir = jiha.archive_dir
            if self.app.config.external_chown_script:
                external_chown(
                    archive_dir,
                    jiha.job.user.system_user_pwent(getpass.getuser()),
                    self.app.config.external_chown_script,
                    "history import archive directory",
                )
            model_store = store.get_import_model_store_for_directory(
                archive_dir, app=self.app, user=user, tag_handler=self.app.tag_handler.create_tag_handler_session()
            )
            job = jiha.job
            with model_store.target_history(default_history=job.history) as new_history:
                jiha.history = new_history
                self.sa_session.flush()
                model_store.perform_import(new_history, job=job, new_history=True)
                # Cleanup.
                if os.path.exists(archive_dir):
                    shutil.rmtree(archive_dir)

        except Exception as e:
            jiha.job.tool_stderr += f"Error cleaning up history import job: {e}"
            self.sa_session.flush()
            raise

        return new_history


class JobExportHistoryArchiveWrapper:
    """
    Class provides support for performing jobs that export a history to an
    archive.
    """

    def __init__(self, app, job_id):
        self.app = app
        self.job_id = job_id
        self.sa_session = self.app.model.context

    def setup_job(self, history, store_directory, include_hidden=False, include_deleted=False, compressed=True):
        """
        Perform setup for job to export a history into an archive.
        """
        # TODO: prevent circular import here...
        from galaxy.celery.tasks import export_history

        app = self.app
        # TODO: parameterize API on include_files here...
        request = SetupHistoryExportJob(
            history_id=history.id,
            job_id=self.job_id,
            store_directory=store_directory,
            include_files=True,
            include_hidden=include_hidden,
            include_deleted=include_deleted,
        )
        if app.config.enable_celery_tasks:
            # symlink files on export, on worker files will tarred up in a dereferenced manner.
            export_history.delay(request=request)
        else:
            export_history(request=request)



"""
Manager and (de)serializer for cloud authorizations (cloudauthzs).
"""

import logging
from typing import Dict

from galaxy import model
from galaxy.exceptions import InternalServerError
from galaxy.managers import (
    base,
    sharable,
)

log = logging.getLogger(__name__)


class CloudAuthzManager(sharable.SharableModelManager):
    model_class = model.CloudAuthz
    foreign_key_name = "cloudauthz"


class CloudAuthzsSerializer(base.ModelSerializer):
    """
    Interface/service object for serializing cloud authorizations (cloudauthzs) into dictionaries.
    """

    model_manager_class = CloudAuthzManager

    def __init__(self, app, **kwargs):
        super().__init__(app, **kwargs)
        self.cloudauthzs_manager = self.manager

        self.default_view = "summary"
        self.add_view(
            "summary",
            [
                "id",
                "model_class",
                "user_id",
                "provider",
                "config",
                "authn_id",
                "last_update",
                "last_activity",
                "create_time",
                "description",
            ],
        )

    def add_serializers(self):
        super().add_serializers()

        # Arguments of the following lambda functions:
        # i  : an instance of galaxy.model.CloudAuthz.
        # k  : serialized dictionary key (e.g., 'model_class', 'provider').
        # **c: a dictionary containing 'trans' and 'user' objects.
        serializers: Dict[str, base.Serializer] = {
            "id": lambda item, key, **context: self.app.security.encode_id(item.id),
            "model_class": lambda item, key, **context: "CloudAuthz",
            "user_id": lambda item, key, **context: self.app.security.encode_id(item.user_id),
            "provider": lambda item, key, **context: str(item.provider),
            "config": lambda item, key, **context: item.config,
            "authn_id": lambda item, key, **context: self.app.security.encode_id(item.authn_id)
            if item.authn_id
            else None,
            "last_update": lambda item, key, **context: str(item.last_update),
            "last_activity": lambda item, key, **context: str(item.last_activity),
            "create_time": lambda item, key, **context: item.create_time.isoformat(),
            "description": lambda item, key, **context: str(item.description),
        }
        self.serializers.update(serializers)


class CloudAuthzsDeserializer(base.ModelDeserializer):
    """
    Service object for validating and deserializing dictionaries that
    update/alter cloudauthz configurations.
    """

    model_manager_class = CloudAuthzManager

    def add_deserializers(self):
        super().add_deserializers()
        self.deserializers.update(
            {
                "authn_id": self.deserialize_and_validate_authn_id,
                "provider": self.default_deserializer,
                "config": self.default_deserializer,
                "description": self.default_deserializer,
            }
        )

    def deserialize_and_validate_authn_id(self, item, key, val, **context):
        """
        Deserializes an authentication ID (authn_id), and asserts if the
        current user can assume that authentication.

        :type  item:    galaxy.model.CloudAuthz
        :param item:    an instance of cloudauthz

        :type  key:     string
        :param key:     `authn_id` attribute of the cloudauthz object (i.e., the `item` param).

        :type  val:     string
        :param val:     the value of `authn_id` attribute of the cloudauthz object (i.e., the `item` param).

        :type  context: dict
        :param context: a dictionary object containing Galaxy `trans`.

        :rtype:         string
        :return:        decoded authentication ID.
        """

        decoded_authn_id = self.app.security.decode_id(val, object_name="authz")

        trans = context.get("trans")
        if trans is None:
            log.debug("Not found expected `trans` when deserializing CloudAuthz.")
            raise InternalServerError

        try:
            trans.app.authnz_manager.can_user_assume_authn(trans, decoded_authn_id)
        except Exception as e:
            raise e

        return decoded_authn_id



import logging
import sqlite3

from sqlalchemy import or_

from galaxy import (
    exceptions,
    model,
)
from galaxy.util.filelock import FileLock

log = logging.getLogger(__name__)

DATABASE_TABLE_NAME = "gxitproxy"


class InteractiveToolSqlite:
    def __init__(self, sqlite_filename, encode_id):
        self.sqlite_filename = sqlite_filename
        self.encode_id = encode_id

    def get(self, key, key_type):
        with FileLock(self.sqlite_filename):
            conn = sqlite3.connect(self.sqlite_filename)
            try:
                c = conn.cursor()
                select = f"""SELECT token, host, port, info
                            FROM {DATABASE_TABLE_NAME}
                            WHERE key=? and key_type=?"""
                c.execute(
                    select,
                    (
                        key,
                        key_type,
                    ),
                )
                try:
                    token, host, port, info = c.fetchone()
                except TypeError:
                    log.warning("get(): invalid key: %s key_type %s", key, key_type)
                    return None
                return dict(key=key, key_type=key_type, token=token, host=host, port=port, info=info)
            finally:
                conn.close()

    def save(self, key, key_type, token, host, port, info=None):
        """
        Writeout a key, key_type, token, value store that is can be used for coordinating
        with external resources.
        """
        assert key, ValueError("A non-zero length key is required.")
        assert key_type, ValueError("A non-zero length key_type is required.")
        assert token, ValueError("A non-zero length token is required.")
        with FileLock(self.sqlite_filename):
            conn = sqlite3.connect(self.sqlite_filename)
            try:
                c = conn.cursor()
                try:
                    # Create table
                    c.execute(
                        """CREATE TABLE %s
                                 (key text,
                                  key_type text,
                                  token text,
                                  host text,
                                  port integer,
                                  info text,
                                  PRIMARY KEY (key, key_type)
                                  )"""
                        % (DATABASE_TABLE_NAME)
                    )
                except Exception:
                    pass
                delete = f"""DELETE FROM {DATABASE_TABLE_NAME} WHERE key=? and key_type=?"""
                c.execute(
                    delete,
                    (
                        key,
                        key_type,
                    ),
                )
                insert = """INSERT INTO %s
                            (key, key_type, token, host, port, info)
                            VALUES (?, ?, ?, ?, ?, ?)""" % (
                    DATABASE_TABLE_NAME
                )
                c.execute(
                    insert,
                    (
                        key,
                        key_type,
                        token,
                        host,
                        port,
                        info,
                    ),
                )
                conn.commit()
            finally:
                conn.close()

    def remove(self, **kwd):
        """
        Remove entry from a key, key_type, token, value store that is can be used for coordinating
        with external resources. Remove entries that match all provided key=values
        """
        assert kwd, ValueError("You must provide some values to key upon")
        delete = f"DELETE FROM {DATABASE_TABLE_NAME} WHERE"
        value_list = []
        for i, (key, value) in enumerate(kwd.items()):
            if i != 0:
                delete += " and"
            delete += f" {key}=?"
            value_list.append(value)
        with FileLock(self.sqlite_filename):
            conn = sqlite3.connect(self.sqlite_filename)
            try:
                c = conn.cursor()
                try:
                    # Delete entry
                    c.execute(delete, tuple(value_list))
                except Exception as e:
                    log.debug("Error removing entry (%s): %s", delete, e)
                conn.commit()
            finally:
                conn.close()

    def save_entry_point(self, entry_point):
        """Convenience method to easily save an entry_point."""
        return self.save(
            self.encode_id(entry_point.id),
            entry_point.__class__.__name__.lower(),
            entry_point.token,
            entry_point.host,
            entry_point.port,
            None,
        )

    def remove_entry_point(self, entry_point):
        """Convenience method to easily remove an entry_point."""
        return self.remove(key=self.encode_id(entry_point.id), key_type=entry_point.__class__.__name__.lower())


class InteractiveToolManager:
    """
    Manager for dealing with InteractiveTools
    """

    def __init__(self, app):
        self.app = app
        self.model = app.model
        self.security = app.security
        self.sa_session = app.model.context
        self.job_manager = app.job_manager
        self.propagator = InteractiveToolSqlite(app.config.interactivetools_map, app.security.encode_id)

    def create_entry_points(self, job, tool, entry_points=None, flush=True):
        entry_points = entry_points or tool.ports
        for entry in entry_points:
            ep = self.model.InteractiveToolEntryPoint(
                job=job,
                tool_port=entry["port"],
                entry_url=entry["url"],
                name=entry["name"],
                requires_domain=entry["requires_domain"],
                short_token=self.app.config.interactivetools_shorten_url,
            )
            self.sa_session.add(ep)
        if flush:
            self.sa_session.flush()

    def configure_entry_point(self, job, tool_port=None, host=None, port=None, protocol=None):
        return self.configure_entry_points(
            job, {tool_port: dict(tool_port=tool_port, host=host, port=port, protocol=protocol)}
        )

    def configure_entry_points(self, job, ports_dict):
        # There can be multiple entry points that reference the same tool port (could have different entry URLs)
        configured = []
        not_configured = []
        for ep in job.interactivetool_entry_points:
            port_dict = ports_dict.get(str(ep.tool_port), None)
            if port_dict is None:
                log.error("Did not find port to assign to InteractiveToolEntryPoint by tool port: %s.", ep.tool_port)
                not_configured.append(ep)
            else:
                ep.host = port_dict["host"]
                ep.port = port_dict["port"]
                ep.protocol = port_dict["protocol"]
                ep.configured = True
                self.sa_session.add(ep)
                self.save_entry_point(ep)
                configured.append(ep)
        if configured:
            self.sa_session.flush()
        return dict(not_configured=not_configured, configured=configured)

    def save_entry_point(self, entry_point):
        """
        Writeout a key, key_type, token, value store that is used validating access
        """
        self.propagator.save_entry_point(entry_point)

    def create_interactivetool(self, job, tool, entry_points):
        # create from initial job
        if job and tool:
            self.create_entry_points(job, tool, entry_points)
        else:
            log.warning(
                "Called InteractiveToolManager.create_interactivetool, but job (%s) or tool (%s) is None", job, tool
            )

    def get_nonterminal_for_user_by_trans(self, trans):
        if trans.user is None and trans.get_galaxy_session() is None:
            return []

        if trans.user:
            jobs = trans.sa_session.query(trans.app.model.Job).filter(trans.app.model.Job.user == trans.user)
        else:
            jobs = trans.sa_session.query(trans.app.model.Job).filter(
                trans.app.model.Job.session_id == trans.get_galaxy_session().id
            )

        def build_and_apply_filters(query, objects, filter_func):
            if objects is not None:
                if isinstance(objects, str):
                    query = query.filter(filter_func(objects))
                elif isinstance(objects, list):
                    t = []
                    for obj in objects:
                        t.append(filter_func(obj))
                    query = query.filter(or_(*t))
            return query

        jobs = build_and_apply_filters(
            jobs, trans.app.model.Job.non_ready_states, lambda s: trans.app.model.Job.state == s
        )
        return trans.sa_session.query(trans.app.model.InteractiveToolEntryPoint).filter(
            trans.app.model.InteractiveToolEntryPoint.job_id.in_([job.id for job in jobs])
        )

    def can_access_job(self, trans, job):
        if job:
            if trans.user is None:
                galaxy_session = trans.get_galaxy_session()
                if galaxy_session is None or job.session_id != galaxy_session.id:
                    return False
            elif job.user != trans.user:
                return False
        else:
            return False
        return True

    def can_access_entry_point(self, trans, entry_point):
        if entry_point:
            return self.can_access_job(trans, entry_point.job)
        return False

    def can_access_entry_points(self, trans, entry_points):
        for ep in entry_points:
            if not self.can_access_entry_point(trans, ep):
                return False
        return True

    def stop(self, trans, entry_point):
        self.remove_entry_point(entry_point)
        job = entry_point.job
        if not job.finished:
            log.debug("Stopping Job: %s for InteractiveToolEntryPoint: %s", job, entry_point)
            job.mark_stopped(trans.app.config.track_jobs_in_database)
            # This self.job_manager.stop(job) does nothing without changing job.state, manually or e.g. with .mark_deleted()
            self.job_manager.stop(job)
            trans.sa_session.add(job)
            trans.sa_session.flush()

    def remove_entry_points(self, entry_points):
        if entry_points:
            for entry_point in entry_points:
                self.remove_entry_point(entry_point, flush=False)
            self.sa_session.flush()

    def remove_entry_point(self, entry_point, flush=True):
        entry_point.deleted = True
        self.sa_session.add(entry_point)
        if flush:
            self.sa_session.flush()
        self.propagator.remove_entry_point(entry_point)

    def target_if_active(self, trans, entry_point):
        if entry_point.active and not entry_point.deleted:
            request_host = trans.request.host
            if not self.app.config.interactivetools_upstream_proxy and self.app.config.interactivetools_proxy_host:
                request_host = self.app.config.interactivetools_proxy_host
            protocol = trans.request.host_url.split("//", 1)[0]
            if entry_point.requires_domain:
                rval = f"{protocol}//{self.get_entry_point_subdomain(trans, entry_point)}.{request_host}/"
                if entry_point.entry_url:
                    rval = "{}/{}".format(rval.rstrip("/"), entry_point.entry_url.lstrip("/"))
            else:
                rval = self.get_entry_point_path(trans, entry_point)
                if not self.app.config.interactivetools_upstream_proxy and self.app.config.interactivetools_proxy_host:
                    rval = f"{protocol}//{request_host}{rval}"
            return rval

    def get_entry_point_subdomain(self, trans, entry_point):
        entry_point_encoded_id = trans.security.encode_id(entry_point.id)
        entry_point_class = entry_point.__class__.__name__.lower()
        entry_point_prefix = self.app.config.interactivetools_prefix
        entry_point_token = entry_point.token
        if self.app.config.interactivetools_shorten_url:
            return f"{entry_point_encoded_id}-{entry_point_token[:10]}.{entry_point_prefix}"
        return f"{entry_point_encoded_id}-{entry_point_token}.{entry_point_class}.{entry_point_prefix}"

    def get_entry_point_path(self, trans, entry_point):
        entry_point_encoded_id = trans.security.encode_id(entry_point.id)
        entry_point_class = entry_point.__class__.__name__.lower()
        entry_point_prefix = self.app.config.interactivetools_prefix
        rval = "/"
        if not entry_point.requires_domain:
            rval = str(self.app.config.interactivetools_base_path).rstrip("/").lstrip("/")
            if self.app.config.interactivetools_shorten_url:
                rval = f"/{rval}/{entry_point_prefix}/{entry_point_encoded_id}/{entry_point.token[:10]}/"
            else:
                rval = f"/{rval}/{entry_point_prefix}/access/{entry_point_class}/{entry_point_encoded_id}/{entry_point.token}/"
        if entry_point.entry_url:
            rval = f"{rval.rstrip('/')}/{entry_point.entry_url.lstrip('/')}"
        rval = "/" + rval.lstrip("/")
        return rval

    def access_entry_point_target(self, trans, entry_point_id):
        entry_point = trans.sa_session.query(model.InteractiveToolEntryPoint).get(entry_point_id)
        if self.app.interactivetool_manager.can_access_entry_point(trans, entry_point):
            if entry_point.active:
                return self.target_if_active(trans, entry_point)
            elif entry_point.deleted:
                raise exceptions.MessageException("InteractiveTool has ended. You will have to start a new one.")
            else:
                raise exceptions.MessageException(
                    "InteractiveTool is not active. If you recently launched this tool it may not be ready yet, please wait a moment and refresh this page."
                )
        raise exceptions.ItemAccessibilityException("You do not have access to this InteractiveTool entry point.")



"""
Tool Input Translation.
"""

import logging

from galaxy.util.bunch import Bunch

log = logging.getLogger(__name__)


class ToolInputTranslator:
    """
    Handles Tool input translation.
    This is used for data source tools

    >>> from galaxy.util import Params, XML
    >>> translator = ToolInputTranslator.from_element(XML(
    ... '''
    ... <request_param_translation>
    ...  <request_param galaxy_name="URL_method" remote_name="URL_method" missing="post" />
    ...  <request_param galaxy_name="URL" remote_name="URL" missing="" >
    ...     <append_param separator="&amp;" first_separator="?" join="=">
    ...         <value name="_export" missing="1" />
    ...         <value name="GALAXY_URL" missing="0" />
    ...     </append_param>
    ...  </request_param>
    ...  <request_param galaxy_name="dbkey" remote_name="db" missing="?" />
    ...  <request_param galaxy_name="organism" remote_name="org" missing="unknown species" />
    ...  <request_param galaxy_name="table" remote_name="hgta_table" missing="unknown table" />
    ...  <request_param galaxy_name="description" remote_name="hgta_regionType" missing="no description" />
    ...  <request_param galaxy_name="data_type" remote_name="hgta_outputType" missing="tabular" >
    ...   <value_translation>
    ...    <value galaxy_value="tabular" remote_value="primaryTable" />
    ...    <value galaxy_value="tabular" remote_value="selectedFields" />
    ...    <value galaxy_value="wig" remote_value="wigData" />
    ...    <value galaxy_value="interval" remote_value="tab" />
    ...    <value galaxy_value="html" remote_value="hyperlinks" />
    ...    <value galaxy_value="fasta" remote_value="sequence" />
    ...   </value_translation>
    ...  </request_param>
    ... </request_param_translation>
    ... '''))
    >>> params = Params({'db':'hg17', 'URL':'URL_value', 'org':'Human', 'hgta_outputType':'primaryTable'})
    >>> translator.translate(params)
    >>> print(sorted(params.__dict__.keys()))
    ['URL', 'URL_method', 'data_type', 'db', 'dbkey', 'description', 'hgta_outputType', 'org', 'organism', 'table']
    >>> params.get('URL', None) in ['URL_value?GALAXY_URL=0&_export=1', 'URL_value?_export=1&GALAXY_URL=0']
    True
    """

    @classmethod
    def from_element(cls, elem):
        """Loads the proper filter by the type attribute of elem"""
        rval = ToolInputTranslator()
        for req_param in elem.findall("request_param"):
            # req_param tags must look like <request_param galaxy_name="dbkey" remote_name="GENOME" missing="" />
            # trans_list = []
            remote_name = req_param.get("remote_name")
            galaxy_name = req_param.get("galaxy_name")
            missing = req_param.get("missing")
            value_trans = {}
            append_param = None

            value_trans_elem = req_param.find("value_translation")
            if value_trans_elem is not None:
                for value_elem in value_trans_elem.findall("value"):
                    remote_value = value_elem.get("remote_value")
                    galaxy_value = value_elem.get("galaxy_value")
                    if None not in [remote_value, galaxy_value]:
                        value_trans[remote_value] = galaxy_value

            append_param_elem = req_param.find("append_param")
            if append_param_elem is not None:
                separator = append_param_elem.get("separator", ",")
                first_separator = append_param_elem.get("first_separator", None)
                join_str = append_param_elem.get("join", "=")
                append_dict = {}
                for value_elem in append_param_elem.findall("value"):
                    value_name = value_elem.get("name")
                    value_missing = value_elem.get("missing")
                    if None not in [value_name, value_missing]:
                        append_dict[value_name] = value_missing
                append_param = Bunch(
                    separator=separator, first_separator=first_separator, join_str=join_str, append_dict=append_dict
                )

            rval.param_trans_dict[remote_name] = Bunch(
                galaxy_name=galaxy_name, missing=missing, value_trans=value_trans, append_param=append_param
            )

        return rval

    def __init__(self):
        self.param_trans_dict = {}

    def translate(self, params):
        """
        update params in-place
        """
        for remote_name, translator in self.param_trans_dict.items():
            galaxy_name = (
                translator.galaxy_name
            )  # NB: if a param by name galaxy_name is provided, it is always thrown away unless galaxy_name == remote_name
            value = params.get(
                remote_name, translator.missing
            )  # get value from input params, or use default value specified in tool config
            if translator.value_trans and value in translator.value_trans:
                value = translator.value_trans[value]
            if translator.append_param:
                for param_name, missing_value in translator.append_param.append_dict.items():
                    param_value = params.get(param_name, missing_value)
                    if translator.append_param.first_separator and translator.append_param.first_separator not in value:
                        sep = translator.append_param.first_separator
                    else:
                        sep = translator.append_param.separator
                    value += f"{sep}{param_name}{translator.append_param.join_str}{param_value}"
            params.update({galaxy_name: value})



import logging
import os
import tempfile

from sqlalchemy import and_

from galaxy import util
from galaxy.tool_shed.repository_type import (
    REPOSITORY_DEPENDENCY_DEFINITION_FILENAME,
    TOOL_DEPENDENCY_DEFINITION_FILENAME,
)
from galaxy.tool_shed.tools.tool_validator import ToolValidator
from galaxy.tool_shed.util import (
    shed_util_common as suc,
    tool_dependency_util,
    tool_util,
)
from galaxy.tool_shed.util.basic_util import (
    remove_dir,
    strip_path,
)
from galaxy.tool_shed.util.hg_util import get_config_from_disk
from galaxy.tool_shed.util.metadata_util import get_updated_changeset_revisions_from_tool_shed
from galaxy.tool_shed.util.repository_util import get_repository_for_dependency_relationship
from galaxy.tool_util.loader_directory import looks_like_a_tool
from galaxy.tool_util.parser.interface import TestCollectionDef
from galaxy.tools.data_manager.manager import DataManager
from galaxy.tools.repositories import ValidationContext
from galaxy.util.tool_shed.common_util import (
    generate_clone_url_for_installed_repository,
    generate_clone_url_for_repository_in_tool_shed,
    remove_protocol_and_user_from_clone_url,
    remove_protocol_from_tool_shed_url,
)
from galaxy.util.tool_shed.xml_util import parse_xml
from galaxy.web import url_for

log = logging.getLogger(__name__)


class MetadataGenerator:
    def __init__(
        self,
        app,
        repository=None,
        changeset_revision=None,
        repository_clone_url=None,
        shed_config_dict=None,
        relative_install_dir=None,
        repository_files_dir=None,
        resetting_all_metadata_on_repository=False,
        updating_installed_repository=False,
        persist=False,
        metadata_dict=None,
        user=None,
    ):
        self.app = app
        self.user = user
        self.repository = repository
        if self.app.name == "galaxy":
            if changeset_revision is None and self.repository is not None:
                self.changeset_revision = self.repository.changeset_revision
            else:
                self.changeset_revision = changeset_revision

            if repository_clone_url is None and self.repository is not None:
                self.repository_clone_url = generate_clone_url_for_installed_repository(self.app, self.repository)
            else:
                self.repository_clone_url = repository_clone_url
            if shed_config_dict is None:
                if self.repository is not None:
                    self.shed_config_dict = self.repository.get_shed_config_dict(self.app)
                else:
                    self.shed_config_dict = {}
            else:
                self.shed_config_dict = shed_config_dict
            if relative_install_dir is None and self.repository is not None:
                tool_path, relative_install_dir = self.repository.get_tool_relative_path(self.app)
            if repository_files_dir is None and self.repository is not None:
                repository_files_dir = self.repository.repo_files_directory(self.app)
            if metadata_dict is None:
                # Shed related tool panel configs are only relevant to Galaxy.
                self.metadata_dict = {"shed_config_filename": self.shed_config_dict.get("config_filename", None)}
            else:
                self.metadata_dict = metadata_dict
        else:
            # We're in the Tool Shed.
            if changeset_revision is None and self.repository is not None:
                self.changeset_revision = self.repository.tip()
            else:
                self.changeset_revision = changeset_revision
            if repository_clone_url is None and self.repository is not None:
                self.repository_clone_url = generate_clone_url_for_repository_in_tool_shed(self.user, self.repository)
            else:
                self.repository_clone_url = repository_clone_url
            if shed_config_dict is None:
                self.shed_config_dict = {}
            else:
                self.shed_config_dict = shed_config_dict
            if relative_install_dir is None and self.repository is not None:
                relative_install_dir = self.repository.repo_path(self.app)
            if repository_files_dir is None and self.repository is not None:
                repository_files_dir = self.repository.repo_path(self.app)
            if metadata_dict is None:
                self.metadata_dict = {}
            else:
                self.metadata_dict = metadata_dict
        self.relative_install_dir = relative_install_dir
        self.repository_files_dir = repository_files_dir
        self.resetting_all_metadata_on_repository = resetting_all_metadata_on_repository
        self.updating_installed_repository = updating_installed_repository
        self.persist = persist
        self.invalid_file_tups = []
        self.sa_session = app.model.session
        self.NOT_TOOL_CONFIGS = [
            suc.DATATYPES_CONFIG_FILENAME,
            REPOSITORY_DEPENDENCY_DEFINITION_FILENAME,
            TOOL_DEPENDENCY_DEFINITION_FILENAME,
            suc.REPOSITORY_DATA_MANAGER_CONFIG_FILENAME,
        ]

    def _generate_data_manager_metadata(
        self, repo_dir, data_manager_config_filename, metadata_dict, shed_config_dict=None
    ):
        """
        Update the received metadata_dict with information from the parsed data_manager_config_filename.
        """
        if data_manager_config_filename is None:
            return metadata_dict
        repo_path = self.repository.repo_path(self.app)
        try:
            # Galaxy Side.
            repo_files_directory = self.repository.repo_files_directory(self.app)
            repo_dir = repo_files_directory
        except AttributeError:
            # Tool Shed side.
            repo_files_directory = repo_path
        relative_data_manager_dir = util.relpath(os.path.split(data_manager_config_filename)[0], repo_dir)
        rel_data_manager_config_filename = os.path.join(
            relative_data_manager_dir, os.path.split(data_manager_config_filename)[1]
        )
        data_managers = {}
        invalid_data_managers = []
        data_manager_metadata = {
            "config_filename": rel_data_manager_config_filename,
            "data_managers": data_managers,
            "invalid_data_managers": invalid_data_managers,
            "error_messages": [],
        }
        metadata_dict["data_manager"] = data_manager_metadata
        tree, error_message = parse_xml(data_manager_config_filename)
        if tree is None:
            # We are not able to load any data managers.
            data_manager_metadata["error_messages"].append(error_message)
            return metadata_dict
        tool_path = None
        if shed_config_dict:
            tool_path = shed_config_dict.get("tool_path", None)
        tools = {}
        for tool in metadata_dict.get("tools", []):
            tool_conf_name = tool["tool_config"]
            if tool_path:
                tool_conf_name = os.path.join(tool_path, tool_conf_name)
            tools[tool_conf_name] = tool
        root = tree.getroot()
        data_manager_tool_path = root.get("tool_path", None)
        if data_manager_tool_path:
            relative_data_manager_dir = os.path.join(relative_data_manager_dir, data_manager_tool_path)
        for i, data_manager_elem in enumerate(root.findall("data_manager")):
            tool_file = data_manager_elem.get("tool_file", None)
            data_manager_id = data_manager_elem.get("id", None)
            if data_manager_id is None:
                log.error(f'Data Manager entry is missing id attribute in "{data_manager_config_filename}".')
                invalid_data_managers.append(
                    {"index": i, "error_message": "Data Manager entry is missing id attribute"}
                )
                continue
            # FIXME: default behavior is to fall back to tool.name.
            data_manager_name = data_manager_elem.get("name", data_manager_id)
            version = data_manager_elem.get("version", DataManager.DEFAULT_VERSION)
            guid = self.generate_guid_for_object(DataManager.GUID_TYPE, data_manager_id, version)
            data_tables = []
            if tool_file is None:
                log.error(f'Data Manager entry is missing tool_file attribute in "{data_manager_config_filename}".')
                invalid_data_managers.append(
                    {"index": i, "error_message": "Data Manager entry is missing tool_file attribute"}
                )
                continue
            else:
                bad_data_table = False
                for data_table_elem in data_manager_elem.findall("data_table"):
                    data_table_name = data_table_elem.get("name", None)
                    if data_table_name is None:
                        log.error(
                            f'Data Manager data_table entry is missing name attribute in "{data_manager_config_filename}".'
                        )
                        invalid_data_managers.append(
                            {"index": i, "error_message": "Data Manager entry is missing name attribute"}
                        )
                        bad_data_table = True
                        break
                    else:
                        data_tables.append(data_table_name)
                if bad_data_table:
                    continue
            data_manager_metadata_tool_file = os.path.normpath(os.path.join(relative_data_manager_dir, tool_file))
            tool_metadata_tool_file = os.path.join(repo_files_directory, data_manager_metadata_tool_file)
            tool = tools.get(tool_metadata_tool_file, None)
            if tool is None:
                log.error(f"Unable to determine tools metadata for '{data_manager_metadata_tool_file}'.")
                invalid_data_managers.append({"index": i, "error_message": "Unable to determine tools metadata"})
                continue
            data_managers[data_manager_id] = {
                "id": data_manager_id,
                "name": data_manager_name,
                "guid": guid,
                "version": version,
                "tool_config_file": data_manager_metadata_tool_file,
                "data_tables": data_tables,
                "tool_guid": tool["guid"],
            }
            log.debug(f"Loaded Data Manager tool_files: {tool_file}")
        return metadata_dict

    def generate_environment_dependency_metadata(self, elem, valid_tool_dependencies_dict):
        """
        The value of env_var_name must match the value of the "set_environment" type
        in the tool config's <requirements> tag set, or the tool dependency will be
        considered an orphan.
        """
        # The value of the received elem looks something like this:
        # <set_environment version="1.0">
        #    <environment_variable name="JAVA_JAR_PATH" action="set_to">$INSTALL_DIR</environment_variable>
        # </set_environment>
        for env_elem in elem:
            # <environment_variable name="JAVA_JAR_PATH" action="set_to">$INSTALL_DIR</environment_variable>
            env_name = env_elem.get("name", None)
            if env_name:
                requirements_dict = dict(name=env_name, type="set_environment")
                if "set_environment" in valid_tool_dependencies_dict:
                    valid_tool_dependencies_dict["set_environment"].append(requirements_dict)
                else:
                    valid_tool_dependencies_dict["set_environment"] = [requirements_dict]
        return valid_tool_dependencies_dict

    def generate_guid_for_object(self, guid_type, obj_id, version):
        tmp_url = remove_protocol_and_user_from_clone_url(self.repository_clone_url)
        return f"{tmp_url}/{guid_type}/{obj_id}/{version}"

    def generate_metadata_for_changeset_revision(self):
        """
        Generate metadata for a repository using its files on disk.  To generate metadata
        for changeset revisions older than the repository tip, the repository will have been
        cloned to a temporary location and updated to a specified changeset revision to access
        that changeset revision's disk files, so the value of self.repository_files_dir will not
        always be self.repository.repo_path( self.app ) (it could be an absolute path to a temporary
        directory containing a clone).  If it is an absolute path, the value of self.relative_install_dir
        must contain repository.repo_path( self.app ).

        The value of self.persist will be True when the installed repository contains a valid
        tool_data_table_conf.xml.sample file, in which case the entries should ultimately be
        persisted to the file referred to by self.app.config.shed_tool_data_table_config.
        """
        if self.shed_config_dict is None:
            self.shed_config_dict = {}
        if self.updating_installed_repository:
            # Keep the original tool shed repository metadata if setting metadata on a repository
            # installed into a local Galaxy instance for which we have pulled updates.
            original_repository_metadata = self.repository.metadata_
        else:
            original_repository_metadata = None
        readme_file_names = _get_readme_file_names(str(self.repository.name))
        if self.app.name == "galaxy":
            # Shed related tool panel configs are only relevant to Galaxy.
            metadata_dict = {"shed_config_filename": self.shed_config_dict.get("config_filename")}
        else:
            metadata_dict = {}
        readme_files = []
        invalid_tool_configs = []
        if self.resetting_all_metadata_on_repository:
            if not self.relative_install_dir:
                raise Exception(
                    "The value of self.repository.repo_path must be set when resetting all metadata on a repository."
                )
            # Keep track of the location where the repository is temporarily cloned so that we can
            # strip the path when setting metadata.  The value of self.repository_files_dir is the
            # full path to the temporary directory to which self.repository was cloned.
            work_dir = self.repository_files_dir
            files_dir = self.repository_files_dir
            # Since we're working from a temporary directory, we can safely copy sample files included
            # in the repository to the repository root.
        else:
            # Use a temporary working directory to copy all sample files.
            work_dir = tempfile.mkdtemp(prefix="tmp-toolshed-gmfcr")
            # All other files are on disk in the repository's repo_path, which is the value of
            # self.relative_install_dir.
            files_dir = self.relative_install_dir
            if self.shed_config_dict.get("tool_path"):
                files_dir = os.path.join(self.shed_config_dict["tool_path"], files_dir)
        # Create ValidationContext to load and validate tools, data tables and datatypes
        with ValidationContext.from_app(app=self.app, work_dir=work_dir) as validation_context:
            tv = ToolValidator(validation_context)
            # Get the relative path to all sample files included in the repository for storage in
            # the repository's metadata.
            sample_file_metadata_paths, sample_file_copy_paths = self.get_sample_files_from_disk(
                repository_files_dir=files_dir,
                tool_path=self.shed_config_dict.get("tool_path"),
                relative_install_dir=self.relative_install_dir,
            )
            if sample_file_metadata_paths:
                metadata_dict["sample_files"] = sample_file_metadata_paths
            # Copy all sample files included in the repository to a single directory location so we
            # can load tools that depend on them.
            data_table_conf_xml_sample_files = []
            for sample_file in sample_file_copy_paths:
                tool_util.copy_sample_file(self.app.config.tool_data_path, sample_file, dest_path=work_dir)
                # If the list of sample files includes a tool_data_table_conf.xml.sample file, load
                # its table elements into memory.
                relative_path, filename = os.path.split(sample_file)
                if filename == "tool_data_table_conf.xml.sample":
                    data_table_conf_xml_sample_files.append(sample_file)

            for data_table_conf_xml_sample_file in data_table_conf_xml_sample_files:
                # We create a new ToolDataTableManager to avoid adding entries to the app-wide
                # tool data tables. This is only used for checking that the data table is valid.
                new_table_elems, error_message = validation_context.tool_data_tables.add_new_entries_from_config_file(
                    config_filename=data_table_conf_xml_sample_file,
                    tool_data_path=work_dir,
                    shed_tool_data_table_config=work_dir,
                    persist=False,
                )
                if error_message:
                    self.invalid_file_tups.append((filename, error_message))
            for root, dirs, files in os.walk(files_dir):
                if root.find(".hg") < 0 and root.find("hgrc") < 0:
                    if ".hg" in dirs:
                        dirs.remove(".hg")
                    for name in files:
                        # See if we have a repository dependencies defined.
                        if name == REPOSITORY_DEPENDENCY_DEFINITION_FILENAME:
                            path_to_repository_dependencies_config = os.path.join(root, name)
                            metadata_dict, error_message = self.generate_repository_dependency_metadata(
                                path_to_repository_dependencies_config, metadata_dict
                            )
                            if error_message:
                                self.invalid_file_tups.append((name, error_message))
                        # See if we have one or more READ_ME files.
                        elif name.lower() in readme_file_names:
                            relative_path_to_readme = self.get_relative_path_to_repository_file(
                                root, name, self.relative_install_dir, work_dir, self.shed_config_dict
                            )
                            readme_files.append(relative_path_to_readme)
                        # See if we have a tool config.
                        elif looks_like_a_tool(os.path.join(root, name), invalid_names=self.NOT_TOOL_CONFIGS):
                            full_path = str(os.path.abspath(os.path.join(root, name)))  # why the str, seems very odd
                            element_tree, error_message = parse_xml(full_path)
                            if element_tree is None:
                                is_tool = False
                            else:
                                element_tree_root = element_tree.getroot()
                                is_tool = element_tree_root.tag == "tool"
                            if is_tool:
                                tool, valid, error_message = tv.load_tool_from_config(
                                    self.app.security.encode_id(self.repository.id), full_path
                                )
                                if tool is None:
                                    if not valid:
                                        invalid_tool_configs.append(name)
                                        self.invalid_file_tups.append((name, error_message))
                                else:
                                    invalid_files_and_errors_tups = tv.check_tool_input_params(
                                        files_dir, name, tool, sample_file_copy_paths
                                    )
                                    can_set_metadata = True
                                    for tup in invalid_files_and_errors_tups:
                                        if name in tup:
                                            can_set_metadata = False
                                            invalid_tool_configs.append(name)
                                            break
                                    if can_set_metadata:
                                        relative_path_to_tool_config = self.get_relative_path_to_repository_file(
                                            root, name, self.relative_install_dir, work_dir, self.shed_config_dict
                                        )
                                        metadata_dict = self.generate_tool_metadata(
                                            relative_path_to_tool_config, tool, metadata_dict
                                        )
                                    else:
                                        for tup in invalid_files_and_errors_tups:
                                            self.invalid_file_tups.append(tup)
        # Handle any data manager entries
        data_manager_config = get_config_from_disk(suc.REPOSITORY_DATA_MANAGER_CONFIG_FILENAME, files_dir)
        metadata_dict = self._generate_data_manager_metadata(
            files_dir, data_manager_config, metadata_dict, shed_config_dict=self.shed_config_dict
        )

        if readme_files:
            metadata_dict["readme_files"] = readme_files
        # This step must be done after metadata for tools has been defined.
        tool_dependencies_config = get_config_from_disk(TOOL_DEPENDENCY_DEFINITION_FILENAME, files_dir)
        if tool_dependencies_config:
            metadata_dict, error_message = self.generate_tool_dependency_metadata(
                tool_dependencies_config, metadata_dict, original_repository_metadata=original_repository_metadata
            )
            if error_message:
                self.invalid_file_tups.append((TOOL_DEPENDENCY_DEFINITION_FILENAME, error_message))
        if invalid_tool_configs:
            metadata_dict["invalid_tools"] = invalid_tool_configs
        self.metadata_dict = metadata_dict
        remove_dir(work_dir)

    def generate_package_dependency_metadata(self, elem, valid_tool_dependencies_dict, invalid_tool_dependencies_dict):
        """
        Generate the metadata for a tool dependencies package defined for a repository.  The
        value of package_name must match the value of the "package" type in the tool config's
        <requirements> tag set.
        """
        # TODO: make this function a class.
        repository_dependency_is_valid = True
        repository_dependency_tup = []
        requirements_dict = {}
        error_message = ""
        package_name = elem.get("name", None)
        package_version = elem.get("version", None)
        if package_name and package_version:
            requirements_dict["name"] = package_name
            requirements_dict["version"] = package_version
            requirements_dict["type"] = "package"
            for sub_elem in elem:
                if sub_elem.tag == "readme":
                    requirements_dict["readme"] = sub_elem.text
                elif sub_elem.tag == "repository":
                    # We have a complex repository dependency.  If the returned value of repository_dependency_is_valid
                    # is True, the tool dependency definition will be set as invalid.  This is currently the only case
                    # where a tool dependency definition is considered invalid.
                    (
                        repository_dependency_tup,
                        repository_dependency_is_valid,
                        error_message,
                    ) = self.handle_repository_elem(repository_elem=sub_elem, only_if_compiling_contained_td=False)
                elif sub_elem.tag == "install":
                    package_install_version = sub_elem.get("version", "1.0")
                    if package_install_version == "1.0":
                        # Complex repository dependencies can be defined within the last <actions> tag set contained in an
                        # <actions_group> tag set.  Comments, <repository> tag sets and <readme> tag sets will be skipped
                        # in tool_dependency_util.parse_package_elem().
                        actions_elem_tuples = tool_dependency_util.parse_package_elem(
                            sub_elem, platform_info_dict=None, include_after_install_actions=False
                        )
                        if actions_elem_tuples:
                            # We now have a list of a single tuple that looks something like:
                            # [(True, <Element 'actions' at 0x104017850>)]
                            actions_elem_tuple = actions_elem_tuples[0]
                            in_actions_group, actions_elem = actions_elem_tuple
                            if in_actions_group:
                                # Since we're inside an <actions_group> tag set, inspect the actions_elem to see if a complex
                                # repository dependency is defined.  By definition, complex repository dependency definitions
                                # contained within the last <actions> tag set within an <actions_group> tag set will have the
                                # value of "only_if_compiling_contained_td" set to True in
                                for action_elem in actions_elem:
                                    if action_elem.tag == "package":
                                        # <package name="libgtextutils" version="0.6">
                                        #    <repository name="package_libgtextutils_0_6" owner="test" prior_installation_required="True" />
                                        # </package>
                                        ae_package_name = action_elem.get("name", None)
                                        ae_package_version = action_elem.get("version", None)
                                        if ae_package_name and ae_package_version:
                                            for sub_action_elem in action_elem:
                                                if sub_action_elem.tag == "repository":
                                                    # We have a complex repository dependency.
                                                    (
                                                        repository_dependency_tup,
                                                        repository_dependency_is_valid,
                                                        error_message,
                                                    ) = self.handle_repository_elem(
                                                        repository_elem=sub_action_elem,
                                                        only_if_compiling_contained_td=True,
                                                    )
                                    elif action_elem.tag == "action":
                                        # <action type="set_environment_for_install">
                                        #    <repository changeset_revision="b107b91b3574" name="package_readline_6_2" owner="devteam" prior_installation_required="True" toolshed="http://localhost:9009">
                                        #        <package name="readline" version="6.2" />
                                        #    </repository>
                                        # </action>
                                        for sub_action_elem in action_elem:
                                            if sub_action_elem.tag == "repository":
                                                # We have a complex repository dependency.
                                                (
                                                    repository_dependency_tup,
                                                    repository_dependency_is_valid,
                                                    error_message,
                                                ) = self.handle_repository_elem(
                                                    repository_elem=sub_action_elem, only_if_compiling_contained_td=True
                                                )
        if requirements_dict:
            dependency_key = f"{package_name}/{package_version}"
            if repository_dependency_is_valid:
                valid_tool_dependencies_dict[dependency_key] = requirements_dict
            else:
                # Append the error message to the requirements_dict.
                requirements_dict["error"] = error_message
                invalid_tool_dependencies_dict[dependency_key] = requirements_dict
        return (
            valid_tool_dependencies_dict,
            invalid_tool_dependencies_dict,
            repository_dependency_tup,
            repository_dependency_is_valid,
            error_message,
        )

    def generate_repository_dependency_metadata(self, repository_dependencies_config, metadata_dict):
        """
        Generate a repository dependencies dictionary based on valid information defined in the received
        repository_dependencies_config.  This method is called from the tool shed as well as from Galaxy.
        """
        # Make sure we're looking at a valid repository_dependencies.xml file.
        tree, error_message = parse_xml(repository_dependencies_config)
        if tree is None:
            xml_is_valid = False
        else:
            root = tree.getroot()
            xml_is_valid = root.tag == "repositories"
        if xml_is_valid:
            invalid_repository_dependencies_dict = dict(description=root.get("description"))
            invalid_repository_dependency_tups = []
            valid_repository_dependencies_dict = dict(description=root.get("description"))
            valid_repository_dependency_tups = []
            for repository_elem in root.findall("repository"):
                repository_dependency_tup, repository_dependency_is_valid, err_msg = self.handle_repository_elem(
                    repository_elem, only_if_compiling_contained_td=False
                )
                if repository_dependency_is_valid:
                    valid_repository_dependency_tups.append(repository_dependency_tup)
                else:
                    # Append the error_message to the repository dependencies tuple.
                    (
                        toolshed,
                        name,
                        owner,
                        changeset_revision,
                        prior_installation_required,
                        only_if_compiling_contained_td,
                    ) = repository_dependency_tup
                    repository_dependency_tup = (
                        toolshed,
                        name,
                        owner,
                        changeset_revision,
                        prior_installation_required,
                        only_if_compiling_contained_td,
                        err_msg,
                    )
                    invalid_repository_dependency_tups.append(repository_dependency_tup)
                    error_message += err_msg
            if invalid_repository_dependency_tups:
                invalid_repository_dependencies_dict["repository_dependencies"] = invalid_repository_dependency_tups
                metadata_dict["invalid_repository_dependencies"] = invalid_repository_dependencies_dict
            if valid_repository_dependency_tups:
                valid_repository_dependencies_dict["repository_dependencies"] = valid_repository_dependency_tups
                metadata_dict["repository_dependencies"] = valid_repository_dependencies_dict
        return metadata_dict, error_message

    def generate_tool_metadata(self, tool_config, tool, metadata_dict):
        """Update the received metadata_dict with changes that have been applied to the received tool."""
        # Generate the guid.
        guid = suc.generate_tool_guid(self.repository_clone_url, tool)
        # Handle tool.requirements.
        tool_requirements = []
        for tool_requirement in tool.requirements:
            name = str(tool_requirement.name)
            tool_type = str(tool_requirement.type)
            version = str(tool_requirement.version) if tool_requirement.version else None
            requirement_dict = dict(name=name, type=tool_type, version=version)
            tool_requirements.append(requirement_dict)
        # Handle tool.tests.
        tool_tests = []
        if tool.tests:
            for ttb in tool.tests:
                required_files = []
                for required_file in ttb.required_files:
                    value, extra = required_file
                    required_files.append(value)
                inputs = []
                for param_name, values in ttb.inputs.items():
                    # Handle improperly defined or strange test parameters and values.
                    if param_name is not None:
                        if values in [None, False]:
                            # An example is the third test in http://testtoolshed.g2.bx.psu.edu/view/devteam/samtools_rmdup
                            # which is defined as:
                            # <test>
                            #    <param name="input1" value="1.bam" ftype="bam" />
                            #    <param name="bam_paired_end_type_selector" value="PE" />
                            #    <param name="force_se" />
                            #    <output name="output1" file="1.bam" ftype="bam" sort="True" />
                            # </test>
                            inputs.append((param_name, values))
                        else:
                            if isinstance(values, TestCollectionDef):
                                # Nested required files are being populated correctly,
                                # not sure we need the value here to be anything else?
                                collection_type = values.collection_type
                                metadata_display_value = f"{collection_type} collection"
                                inputs.append((param_name, metadata_display_value))
                            else:
                                try:
                                    if len(values) == 1:
                                        inputs.append((param_name, values[0]))
                                        continue
                                except TypeError:
                                    log.exception(
                                        'Expected a list of values for tool "%s" parameter "%s", got %s: %s',
                                        tool.id,
                                        param_name,
                                        type(values),
                                        values,
                                    )
                                inputs.append((param_name, values))
                outputs = []
                for output in ttb.outputs:
                    name, file_name, extra = output
                    outputs.append((name, strip_path(file_name) if file_name else None))
                    if file_name not in required_files and file_name is not None:
                        required_files.append(file_name)
                test_dict = dict(name=str(ttb.name), required_files=required_files, inputs=inputs, outputs=outputs)
                tool_tests.append(test_dict)
        # Determine if the tool should be loaded into the tool panel.  Examples of valid tools that
        # should not be displayed in the tool panel are datatypes converters and DataManager tools
        # (which are of type 'manage_data').
        add_to_tool_panel_attribute = self._set_add_to_tool_panel_attribute_for_tool(tool)
        tool_dict = dict(
            id=tool.id,
            guid=guid,
            name=tool.name,
            version=tool.version,
            profile=tool.profile,
            description=tool.description,
            version_string_cmd=tool.version_string_cmd,
            tool_config=tool_config,
            tool_type=tool.tool_type,
            requirements=tool_requirements,
            tests=tool_tests,
            add_to_tool_panel=add_to_tool_panel_attribute,
        )
        if "tools" in metadata_dict:
            metadata_dict["tools"].append(tool_dict)
        else:
            metadata_dict["tools"] = [tool_dict]
        return metadata_dict

    def generate_tool_dependency_metadata(
        self, tool_dependencies_config, metadata_dict, original_repository_metadata=None
    ):
        """
        If the combination of name, version and type of each element is defined in the <requirement> tag for
        at least one tool in self.repository, then update the received metadata_dict with information from the
        parsed tool_dependencies_config.
        """
        error_message = ""
        if original_repository_metadata:
            # Keep a copy of the original tool dependencies dictionary and the list of tool
            # dictionaries in the metadata.
            original_valid_tool_dependencies_dict = original_repository_metadata.get("tool_dependencies", None)
        else:
            original_valid_tool_dependencies_dict = None
        tree, error_message = parse_xml(tool_dependencies_config)
        if tree is None:
            return metadata_dict, error_message
        root = tree.getroot()

        class RecurserValueStore:
            pass

        rvs = RecurserValueStore()
        rvs.valid_tool_dependencies_dict = {}
        rvs.invalid_tool_dependencies_dict = {}
        valid_repository_dependency_tups = []
        invalid_repository_dependency_tups = []
        description = root.get("description")

        def _check_elem_for_dep(elems):
            error_messages = []
            for elem in elems:
                if elem.tag == "package":
                    (
                        rvs.valid_tool_dependencies_dict,
                        rvs.invalid_tool_dependencies_dict,
                        repository_dependency_tup,
                        repository_dependency_is_valid,
                        message,
                    ) = self.generate_package_dependency_metadata(
                        elem, rvs.valid_tool_dependencies_dict, rvs.invalid_tool_dependencies_dict
                    )
                    if repository_dependency_is_valid:
                        if (
                            repository_dependency_tup
                            and repository_dependency_tup not in valid_repository_dependency_tups
                        ):
                            # We have a valid complex repository dependency.
                            valid_repository_dependency_tups.append(repository_dependency_tup)
                    else:
                        if (
                            repository_dependency_tup
                            and repository_dependency_tup not in invalid_repository_dependency_tups
                        ):
                            # We have an invalid complex repository dependency, so mark the tool dependency as invalid.
                            # Append the error message to the invalid repository dependency tuple.
                            (
                                toolshed,
                                name,
                                owner,
                                changeset_revision,
                                prior_installation_required,
                                only_if_compiling_contained_td,
                            ) = repository_dependency_tup
                            repository_dependency_tup = (
                                toolshed,
                                name,
                                owner,
                                changeset_revision,
                                prior_installation_required,
                                only_if_compiling_contained_td,
                                message,
                            )
                            invalid_repository_dependency_tups.append(repository_dependency_tup)
                            error_messages.append(f"{error_message}  {message}")
                elif elem.tag == "set_environment":
                    rvs.valid_tool_dependencies_dict = self.generate_environment_dependency_metadata(
                        elem, rvs.valid_tool_dependencies_dict
                    )
                error_messages += _check_elem_for_dep(elem)
            return error_messages

        error_message = "\n".join([error_message] + _check_elem_for_dep(root))
        if rvs.valid_tool_dependencies_dict:
            if original_valid_tool_dependencies_dict:
                # We're generating metadata on an update pulled to a tool shed repository installed
                # into a Galaxy instance, so handle changes to tool dependencies appropriately.
                irm = self.app.installed_repository_manager
                (
                    updated_tool_dependency_names,
                    deleted_tool_dependency_names,
                ) = irm.handle_existing_tool_dependencies_that_changed_in_update(
                    self.repository, original_valid_tool_dependencies_dict, rvs.valid_tool_dependencies_dict
                )
            metadata_dict["tool_dependencies"] = rvs.valid_tool_dependencies_dict
        if rvs.invalid_tool_dependencies_dict:
            metadata_dict["invalid_tool_dependencies"] = rvs.invalid_tool_dependencies_dict
        if valid_repository_dependency_tups:
            metadata_dict = self.update_repository_dependencies_metadata(
                metadata=metadata_dict,
                repository_dependency_tups=valid_repository_dependency_tups,
                is_valid=True,
                description=description,
            )
        if invalid_repository_dependency_tups:
            metadata_dict = self.update_repository_dependencies_metadata(
                metadata=metadata_dict,
                repository_dependency_tups=invalid_repository_dependency_tups,
                is_valid=False,
                description=description,
            )
        return metadata_dict, error_message

    def get_invalid_file_tups(self):
        return self.invalid_file_tups

    def get_metadata_dict(self):
        return self.metadata_dict

    def get_relative_path_to_repository_file(self, root, name, relative_install_dir, work_dir, shed_config_dict):
        if self.resetting_all_metadata_on_repository:
            full_path_to_file = os.path.join(root, name)
            stripped_path_to_file = full_path_to_file.replace(work_dir, "")
            if stripped_path_to_file.startswith("/"):
                stripped_path_to_file = stripped_path_to_file[1:]
            relative_path_to_file = os.path.join(relative_install_dir, stripped_path_to_file)
        else:
            relative_path_to_file = os.path.join(root, name)
            if (
                relative_install_dir
                and shed_config_dict.get("tool_path")
                and relative_path_to_file.startswith(
                    os.path.join(shed_config_dict.get("tool_path"), relative_install_dir)
                )
            ):
                relative_path_to_file = relative_path_to_file[len(shed_config_dict.get("tool_path")) + 1 :]
        return relative_path_to_file

    def get_sample_files_from_disk(self, repository_files_dir, tool_path=None, relative_install_dir=None):
        work_dir = ""
        if self.resetting_all_metadata_on_repository:
            # Keep track of the location where the repository is temporarily cloned so that we can strip
            # it when setting metadata.
            work_dir = repository_files_dir
        sample_file_metadata_paths = []
        sample_file_copy_paths = []
        for root, _dirs, files in os.walk(repository_files_dir):
            if root.find(".hg") < 0:
                for name in files:
                    if name.endswith(".sample"):
                        if self.resetting_all_metadata_on_repository:
                            full_path_to_sample_file = os.path.join(root, name)
                            stripped_path_to_sample_file = full_path_to_sample_file.replace(work_dir, "")
                            if stripped_path_to_sample_file.startswith("/"):
                                stripped_path_to_sample_file = stripped_path_to_sample_file[1:]
                            relative_path_to_sample_file = os.path.join(
                                relative_install_dir, stripped_path_to_sample_file
                            )
                            if os.path.exists(relative_path_to_sample_file):
                                sample_file_copy_paths.append(relative_path_to_sample_file)
                            else:
                                sample_file_copy_paths.append(full_path_to_sample_file)
                        else:
                            relative_path_to_sample_file = os.path.join(root, name)
                            sample_file_copy_paths.append(relative_path_to_sample_file)
                            if tool_path and relative_install_dir:
                                if relative_path_to_sample_file.startswith(
                                    os.path.join(tool_path, relative_install_dir)
                                ):
                                    relative_path_to_sample_file = relative_path_to_sample_file[len(tool_path) + 1 :]
                        sample_file_metadata_paths.append(relative_path_to_sample_file)
        return sample_file_metadata_paths, sample_file_copy_paths

    def handle_repository_elem(self, repository_elem, only_if_compiling_contained_td=False):
        """
        Process the received repository_elem which is a <repository> tag either from a
        repository_dependencies.xml file or a tool_dependencies.xml file.  If the former,
        we're generating repository dependencies metadata for a repository in the Tool Shed.
        If the latter, we're generating package dependency metadata within Galaxy or the
        Tool Shed.
        """
        is_valid = True
        error_message = ""
        toolshed = repository_elem.get("toolshed", None)
        name = repository_elem.get("name", None)
        owner = repository_elem.get("owner", None)
        changeset_revision = repository_elem.get("changeset_revision", None)
        prior_installation_required = str(repository_elem.get("prior_installation_required", False))
        repository_dependency_tup = [
            toolshed,
            name,
            owner,
            changeset_revision,
            prior_installation_required,
            str(only_if_compiling_contained_td),
        ]
        if self.app.name == "galaxy":
            if self.updating_installed_repository:
                pass
            else:
                # We're installing a repository into Galaxy, so make sure its contained repository
                # dependency definition is valid.
                if toolshed is None or name is None or owner is None or changeset_revision is None:
                    # Several packages exist in the Tool Shed that contain invalid repository
                    # definitions, but will still install. We will report these errors to the
                    # installing user. Previously, we would:
                    # Raise an exception here instead of returning an error_message to keep the
                    # installation from proceeding.  Reaching here implies a bug in the Tool Shed
                    # framework.
                    error_message = "Installation encountered an invalid repository dependency definition:\n"
                    error_message += util.xml_to_string(repository_elem, pretty=True)
                    log.error(error_message)
                    return repository_dependency_tup, False, error_message
        if not toolshed:
            # Default to the current tool shed.
            toolshed = str(url_for("/", qualified=True)).rstrip("/")
            repository_dependency_tup[0] = toolshed
        toolshed = remove_protocol_from_tool_shed_url(toolshed)
        if self.app.name == "galaxy":
            # We're in Galaxy.  We reach here when we're generating the metadata for a tool
            # dependencies package defined for a repository or when we're generating metadata
            # for an installed repository.  See if we can locate the installed repository via
            # the changeset_revision defined in the repository_elem (it may be outdated).  If
            # we're successful in locating an installed repository with the attributes defined
            # in the repository_elem, we know it is valid.
            repository = get_repository_for_dependency_relationship(self.app, toolshed, name, owner, changeset_revision)
            if repository:
                return repository_dependency_tup, is_valid, error_message
            else:
                # Send a request to the tool shed to retrieve appropriate additional changeset
                # revisions with which the repository
                # may have been installed.
                text = get_updated_changeset_revisions_from_tool_shed(
                    self.app, toolshed, name, owner, changeset_revision
                )
                if text:
                    updated_changeset_revisions = util.listify(text)
                    for updated_changeset_revision in updated_changeset_revisions:
                        repository = get_repository_for_dependency_relationship(
                            self.app, toolshed, name, owner, updated_changeset_revision
                        )
                        if repository:
                            return repository_dependency_tup, is_valid, error_message
                        if self.updating_installed_repository:
                            # The repository dependency was included in an update to the installed
                            # repository, so it will not yet be installed.  Return the tuple for later
                            # installation.
                            return repository_dependency_tup, is_valid, error_message
                if self.updating_installed_repository:
                    # The repository dependency was included in an update to the installed repository,
                    # so it will not yet be installed.  Return the tuple for later installation.
                    return repository_dependency_tup, is_valid, error_message
                # Don't generate an error message for missing repository dependencies that are required
                # only if compiling the dependent repository's tool dependency.
                if not only_if_compiling_contained_td:
                    # We'll currently default to setting the repository dependency definition as invalid
                    # if an installed repository cannot be found.  This may not be ideal because the tool
                    # shed may have simply been inaccessible when metadata was being generated for the
                    # installed tool shed repository.
                    error_message = (
                        "Ignoring invalid repository dependency definition for tool shed %s, name %s, owner %s, "
                        % (toolshed, name, owner)
                    )
                    error_message += f"changeset revision {changeset_revision}."
                    log.debug(error_message)
                    is_valid = False
                    return repository_dependency_tup, is_valid, error_message
        else:
            # We're in the tool shed.
            if suc.tool_shed_is_this_tool_shed(toolshed):
                try:
                    user = (
                        self.sa_session.query(self.app.model.User)
                        .filter(self.app.model.User.table.c.username == owner)
                        .one()
                    )
                except Exception:
                    error_message = (
                        "Ignoring repository dependency definition for tool shed %s, name %s, owner %s, "
                        % (toolshed, name, owner)
                    )
                    error_message += f"changeset revision {changeset_revision} because the owner is invalid."
                    log.debug(error_message)
                    is_valid = False
                    return repository_dependency_tup, is_valid, error_message
                try:
                    repository = (
                        self.sa_session.query(self.app.model.Repository)
                        .filter(
                            and_(
                                self.app.model.Repository.table.c.name == name,
                                self.app.model.Repository.table.c.user_id == user.id,
                            )
                        )
                        .one()
                    )
                except Exception:
                    error_message = (
                        "Ignoring repository dependency definition for tool shed %s, name %s, owner %s, "
                        % (toolshed, name, owner)
                    )
                    error_message += f"changeset revision {changeset_revision} because the name is invalid.  "
                    log.debug(error_message)
                    is_valid = False
                    return repository_dependency_tup, is_valid, error_message
                repo = repository.hg_repo

                # The received changeset_revision may be None since defining it in the dependency definition is optional.
                # If this is the case, the default will be to set its value to the repository dependency tip revision.
                # This probably occurs only when handling circular dependency definitions.
                tip_ctx = repo[repo.changelog.tip()]
                # Make sure the repo.changlog includes at least 1 revision.
                if changeset_revision is None and tip_ctx.rev() >= 0:
                    changeset_revision = str(tip_ctx)
                    repository_dependency_tup = [
                        toolshed,
                        name,
                        owner,
                        changeset_revision,
                        prior_installation_required,
                        str(only_if_compiling_contained_td),
                    ]
                    return repository_dependency_tup, is_valid, error_message
                else:
                    # Find the specified changeset revision in the repository's changelog to see if it's valid.
                    found = False
                    for changeset in repo.changelog:
                        changeset_hash = str(repo[changeset])
                        if changeset_hash == changeset_revision:
                            found = True
                            break
                    if not found:
                        error_message = (
                            "Ignoring repository dependency definition for tool shed %s, name %s, owner %s, "
                            % (toolshed, name, owner)
                        )
                        error_message += (
                            f"changeset revision {changeset_revision} because the changeset revision is invalid.  "
                        )
                        log.debug(error_message)
                        is_valid = False
                        return repository_dependency_tup, is_valid, error_message
            else:
                # Repository dependencies are currently supported within a single tool shed.
                error_message = (
                    "Repository dependencies are currently supported only within the same tool shed.  Ignoring "
                )
                error_message += (
                    "repository dependency definition  for tool shed %s, name %s, owner %s, changeset revision %s.  "
                    % (toolshed, name, owner, changeset_revision)
                )
                log.debug(error_message)
                is_valid = False
                return repository_dependency_tup, is_valid, error_message
        return repository_dependency_tup, is_valid, error_message

    def _set_add_to_tool_panel_attribute_for_tool(self, tool):
        """
        Determine if a tool should be loaded into the Galaxy tool panel.  Examples of valid tools that
        should not be displayed in the tool panel are DataManager tools.
        """
        if hasattr(tool, "tool_type"):
            if tool.tool_type in ["manage_data"]:
                # We have a DataManager tool.
                return False
        return True

    def set_changeset_revision(self, changeset_revision):
        self.changeset_revision = changeset_revision

    def set_relative_install_dir(self, relative_install_dir):
        self.relative_install_dir = relative_install_dir

    def set_repository(self, repository, relative_install_dir=None, changeset_revision=None):
        self.repository = repository
        # Shed related tool panel configs are only relevant to Galaxy.
        if self.app.name == "galaxy":
            if relative_install_dir is None and self.repository is not None:
                tool_path, relative_install_dir = self.repository.get_tool_relative_path(self.app)
            if changeset_revision is None and self.repository is not None:
                self.set_changeset_revision(self.repository.changeset_revision)
            else:
                self.set_changeset_revision(changeset_revision)
            self.shed_config_dict = repository.get_shed_config_dict(self.app)
            self.metadata_dict = {"shed_config_filename": self.shed_config_dict.get("config_filename", None)}
        else:
            if relative_install_dir is None and self.repository is not None:
                relative_install_dir = repository.repo_path(self.app)
            if changeset_revision is None and self.repository is not None:
                self.set_changeset_revision(self.repository.tip())
            else:
                self.set_changeset_revision(changeset_revision)
            self.shed_config_dict = {}
            self.metadata_dict = {}
        self.set_relative_install_dir(relative_install_dir)
        self.set_repository_files_dir()
        self.resetting_all_metadata_on_repository = False
        self.updating_installed_repository = False
        self.persist = False
        self.invalid_file_tups = []

    def set_repository_clone_url(self, repository_clone_url):
        self.repository_clone_url = repository_clone_url

    def set_repository_files_dir(self, repository_files_dir=None):
        self.repository_files_dir = repository_files_dir

    def update_repository_dependencies_metadata(self, metadata, repository_dependency_tups, is_valid, description):
        if is_valid:
            repository_dependencies_dict = metadata.get("repository_dependencies", None)
        else:
            repository_dependencies_dict = metadata.get("invalid_repository_dependencies", None)
        for repository_dependency_tup in repository_dependency_tups:
            if repository_dependencies_dict:
                repository_dependencies = repository_dependencies_dict.get("repository_dependencies", [])
                for repository_dependency_tup in repository_dependency_tups:
                    if repository_dependency_tup not in repository_dependencies:
                        repository_dependencies.append(repository_dependency_tup)
                repository_dependencies_dict["repository_dependencies"] = repository_dependencies
            else:
                repository_dependencies_dict = dict(
                    description=description, repository_dependencies=repository_dependency_tups
                )
        if repository_dependencies_dict:
            if is_valid:
                metadata["repository_dependencies"] = repository_dependencies_dict
            else:
                metadata["invalid_repository_dependencies"] = repository_dependencies_dict
        return metadata


def _get_readme_file_names(repository_name):
    """Return a list of file names that will be categorized as README files for the received repository_name."""
    readme_files = ["readme", "read_me", "install"]
    valid_filenames = [f"{f}.txt" for f in readme_files]
    valid_filenames.extend([f"{f}.rst" for f in readme_files])
    valid_filenames.extend(readme_files)
    valid_filenames.append(f"{repository_name}.txt")
    valid_filenames.append(f"{repository_name}.rst")
    return valid_filenames
